/* lunar.h
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   LUNAR logic header file
 *
 * Copyright (C) 2001-2003 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */


// ----------------------------------------------------------------------
#ifndef LUNAR_H
#define LUNAR_H

#include "platform.h"

// all times are in microseconds

#define LUNAR_MIN_RING               1
#define LUNAR_MAX_RING               3

#define LUNAR_TIMEOUT_PER_RING   50000  // 20 msec RTT seems fair
#define LUNAR_RETRY_WAIT        150000  // to wait after reaching max_radius
#define LUNAR_MAX_RETRIES            (LUNAR_MAX_RING + 3)

#define LUNAR_RRET             3000000  // route re-establishment time, in usec
#define LUNAR_RRET_JITTER        50000

#define LUNAR_IP_VERIFY_MAX          3  // if three probes fail, we claim the address
#define LUNAR_IP_PROBE_RETRY    500000  // interval before probing again

#define LUNAR_BRCAST_DELAY       30000  // grace period for path setup before blasting

//The following lines are added by me***********************************
#define CHECK_TIME               250000
#define CHECK_PERIOD             500000
#define DATA_TIMEOUT             1500000
#define RESV_TIMEOUT             300000
//**********************************************************************

// ----------------------------------------------------------------------
// implementation constants/limits (not related to the LUNAR protocol):

#define LUNAR_IP_SUBNET  ((192<<24)|(168<<16)|(42<<8))
#define LUNAR_IP_NETMASK (0xffffff00)
#define LUNAR_IP_BCAST   (LUNAR_IP_SUBNET | ~LUNAR_IP_NETMASK)



// -------------------IMPLEMENTATION SPECIFIC--------------------------
#define LUNAR_MAX_PEERS            32  // max number of active IP peers
#define LUNAR_MAX_FLOWS            64  // max number of active flows at a node

// #define PEER_ARP_PENDING 0x01
#define PEER_BIDIR       0x01
#define PEER_HAS_HID     0x02
#define PEER_BCTREE      0x04


struct uc_pending_s {
	sapfsel_t caller;
	sapfsel_t backsel; // local selector for backw data
	struct in_addr ia;
	char is_bidir;
};

struct uc_transit_pending_s {
	sapfsel_t caller;
	sapfsel_t mybacksel;
	sapfsel_t replysel;
	sapfsel_t backsel;
	unsigned char replyeth[ETH_ALEN];
	unsigned char backeth[ETH_ALEN];
	char is_bidir;
};

struct bcast_pending_s {
	sapfsel_t caller;
	sapfsel_t fwd_sel;
	struct in_addr ia;
};

struct bc_fanout_s {
	sapfsel_t local;
	sapfsel_t forward;
};

// ----------------------------------------------------------------------

struct sapfdest_s {
	int ifnr;
	struct sockaddr dst;
	sapfsel_t sel;
};


#ifndef NS_PORT
struct peer_s {
#else
class peer_s {
    public:  
#endif
	sapfsel_t callback;	      // local handler to remove
	sapfsel_t fwdsel;	      // where to forward to
	sapfsel_t backsel;            // where to backward to
	unsigned char hid[NETBOX_HID_LEN]; // host id
	unsigned long hid_timeout;
	struct in_addr ia;            // peer's ip address, network order
	netbox_timer_t timer;	      // for addr garbage collection
	short int used;               // 0=unused, 1...=resolving, -1=resolved
	unsigned char hops;	      // distance
	unsigned char flags;	      // see below for flag values
        //uint8_t seqn;                 //sequence number for this flow
        double bw;                    //bandwidth requirement
        //short int flow_id;            //flow id 
};

#ifndef NS_PORT
struct flow_s {
#else
class flow_s {
    public:
#endif
        int src;                      //source address
        int dst;                      //destination address
        //short int seqn;               //sequence number for this flow
        netbox_timer_t timer;
        short int status;             //0=unused, 1=pre-reserve, 2=reserved, 3=active
        double bw;                    //the reserved bandwidth
        double last_bw;              
        //double update;                //update time (in fact, it's only used in dest node)
        //short int data;
        //netbox_timer_t timer;         //timer
};
#endif // LUNAR_H

// eof
