/* klunar.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   Linux kernel module wrapper for the LUNAR logic
 *
 * Copyright (C) 2001-2004 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *   2004-01-09 moved generic init stuff into src/lunar_init.c,
 *              added mackill parameter parsing
 *   2004-01-05 the sources are now managed by a CVS
 *   2003-11-20 adjusted debuglevels of printk messages, init/fin of
 *              module now gets announced
 */

#define __KERNEL__
#define __OPTIMIZE__
#define MODULE

#include "../include/platform.h"
#include "../include/version.h"
//********for /proc************
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
//#include <stdlib.h>

//#include "linux/modversions.h"

#include <net/neighbour.h>
#ifndef CONFIG_ARM
 
# include <net/arp.h>
#else
extern struct neigh_table arp_tbl;
# include <endian.h>
#endif

#include "../include/sapf.h"
#include "../include/xrp.h"
#include "../include/netbox.h"
#include "../include/lunar.h"
#include "../include/sapf_decl.h"
#include "../include/xrp_decl.h"
#include "../include/netbox_decl.h"
#include "../include/lunar_decl.h"

int conf_min_ring = LUNAR_MIN_RING;
int conf_max_ring = LUNAR_MAX_RING;
int conf_timeout_per_ring = LUNAR_TIMEOUT_PER_RING;
int conf_max_retries = LUNAR_MAX_RETRIES;
static int conf_fri = LUNAR_RRET;           // forced rediscovery interval
int conf_fri_jitter = LUNAR_RRET_JITTER;
int conf_rreq_gap = 0;
int conf_data_gap = 0;

//the following is added by me************************
//this is like the bind function in ns2
int check_time = CHECK_TIME;
int check_period = CHECK_PERIOD;
int data_timeout = DATA_TIMEOUT;
int resv_timeout = RESV_TIMEOUT;

unsigned long Trts = (unsigned long)(248*1000000/bRate);
unsigned long Tcts = (unsigned long)(240*1000000/bRate);
unsigned long Tack = (unsigned long)(240*1000000/bRate);
#define DSSS_SIFSTime 10    //10us 
#define DSSS_SlotTime 20    //20us

//*****************************************************

//******the following lines are for /proc**************
#define PROCFS_MAX_SIZE 100
#define PROCFS_NAME     "pkt_rate"

static struct proc_dir_entry *Our_Proc_File;
static char procfs_buffer[PROCFS_MAX_SIZE]="40X6000";
static unsigned long procfs_buffer_size = 100;
//char buffer[10];
char buffer[100];

//*****************************************************
int xrp_param_len[XRP_PARTYPE_COUNT] =
{       // XRP_PARTYPE_UNKNOWN:
	0,
	// XRP_PARTYPE_SEL:
	sizeof(struct xrp_common_s) + sizeof(sapfsel_t),
	// XRP_PARTYPE_IPv4
	sizeof(struct xrp_common_s) + sizeof(uint32_t),
	// XRP_PARTYPE_IPv6
	sizeof(struct xrp_common_s) + 16,
	// XRP_PARTYPE_SEL_ETH
	sizeof(struct xrp_common_s) + sizeof(sapfsel_t) + ETH_ALEN,
	// XRP_PARTYPE_SEL_UDP
	sizeof(struct xrp_common_s) + sizeof(sapfsel_t) + sizeof(uint32_t)
						      + sizeof(short),
	sizeof(struct xrp_common_s) + sizeof(double),
	// XRP_PARTYPE_HID128
	sizeof(struct xrp_common_s) + 16
};

#define MAX_MACKILL 3
static int mackill_cnt;
static unsigned char mackill_list[MAX_MACKILL][ETH_ALEN];

struct peer_s peers[LUNAR_MAX_PEERS];

//The following line is added by me***********************
struct flow_s flows[LUNAR_MAX_FLOWS];    //LUNAR_MAX_FLOWS is defined in lunar.h
//double csn_bw = Bmax, local_bw_cs= Bmax;
long csn_bw = Bmax, local_bw_cs=Bmax;
//********************************************************

#include "../src/lunar_init.c"
#include "knetbox.c"

// ----------------------------------------------------------------------

MODULE_AUTHOR("christian.tschudin@unibas.ch");
MODULE_DESCRIPTION("LUNAR kernel module (" LUNAR_VERSION "), 2002-2004");
MODULE_LICENSE("GPL");

MODULE_PARM(basedevname, "s");
MODULE_PARM_DESC(basedevname, "name of base device (default: first device "
		          "with wireless extensions)");

MODULE_PARM(debuglevel, "i");
MODULE_PARM_DESC(debuglevel, "debug level 0-10");

MODULE_PARM(forwarding, "s");
MODULE_PARM_DESC(forwarding, "default gatewaying (no, yes)");

MODULE_PARM(mackill, "s");
MODULE_PARM_DESC(mackill, "suppress packets from the given MAC address/es");
static char *mackill;

// ----------------------------------------------------------------------


static int klunar_netdev_event(struct notifier_block *this,
			       unsigned long event, void *ptr);
static struct notifier_block klunar_notifier = {
	notifier_call: klunar_netdev_event,
};

static int klunar_netdev_event(struct notifier_block *this,
                              unsigned long event, void *ptr)
{
        struct net_device *dev = (struct net_device *) ptr;

	if (!dev)
		return NOTIFY_DONE;

        DEBUG(10, printk("klunar notifier event %08lx %s\n", event, dev->name));

        switch (event) {
        case NETDEV_REGISTER:
        case NETDEV_UP:
		DEBUG(10, printk("notifier up %s\n", dev->name));
		if (basedev)
			break;
		if (basedevname) {
			struct net_device *d = dev_get_by_name(basedevname);
			if (!d)
				break;
			dev_put(d);
			if (d->ifindex != dev->ifindex)
				break;
			if (!strcmp(basedevname, dev->name)) {
				DEBUG(1, printk("klunar - new interface %s cannot be the lunar device\n", basedevname));
				break;
			}
		} else {
			if (!dev->get_wireless_stats)
				break;
		}
		DEBUG(1, printk("klunar - new (wireless) interface %s found\n",
				 dev->name));
		basedev = dev;
		netbox_addL2if(dev->ifindex, myhandle);
                break;
        case NETDEV_UNREGISTER:
        case NETDEV_DOWN:
		DEBUG(10, printk("notifier down %s\n", dev->name));
                if (dev == basedev) {
                        DEBUG(1, printk("klunar - (wireless) interface %s went away\n",
					 dev->name));
			basedev = 0;
			netbox_dropL2if(dev->ifindex);
                }
                break;

        default:
                break;
        };

        return NOTIFY_DONE;
}

// ----------------------------------------------------------------------

#include "../src/lunar.c"
#include "../src/lunar_dhcp.c"
#include "../src/xrp.c"

// ----------------------------------------------------------------------

// The following function is added by me*********************************
/*static void init_check_bw()       //start to periodically check the bandwidth
{
  check_bandwidth_cs((void*)0);
}*/
//**********************************************************************

static int parse_mackill()
{
	char *pa[MAX_MACKILL], *cp;
	int i, j; // , ia[ETH_ALEN];

	DEBUG(10, printk("mackill parse: %s\n", mackill));

	while (mackill_cnt < MAX_MACKILL) {
		pa[mackill_cnt] = strtok(mackill_cnt ? NULL : mackill, ",");
		if (!pa[mackill_cnt])
			break;
		mackill_cnt++;
	}
	for (i = 0; i < mackill_cnt; i++) {
		// lnx kernel bug in 2.4.X: sscanf format "%x" does not work ....
		cp = pa[i];
		for (j = 0; j < ETH_ALEN; j++, cp++) {
			mackill_list[i][j] = 0;
			for (; isxdigit(*cp); cp++) {
				mackill_list[i][j] =
					(mackill_list[i][j] << 4) |
					(*cp <= '9' ? (*cp - '0') :
					           ((*cp & 0x07) + 9 ));
			}
			if (*cp && *cp != ':') break;
		}
		if ( j != ETH_ALEN) {
			DEBUG(0, printk("mackill: error in MAC addr %s\n", pa[i]));
			mackill_cnt--;
			return -1;
		}

		DEBUG(5, printk("mackill +%s\n",
				netbox_eth2str(mackill_list[i])));
	}
	return 0;
}

// ----------------------------------------------------------------------

static int klunar_open(struct net_device *dev)
{
	netif_start_queue(dev);
        return 0;
}


static int klunar_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
  //this function is an analog to selnet::ipinput()
        struct net_device_stats *stats = (struct net_device_stats *) dev->priv;
	sapfsel_t sel;
	sapfpkt pkt;
        //following added by me**************************************
        struct in_addr dst;
        int length;          //length of the data packet
        unsigned long rbw;          //required bandwidth
        
       
        //***********************************************************

        DEBUG(10, printk("klunar start_xmit (%d bytes)\n", skb->len));

        switch (ntohs(*(unsigned short *) (skb->data + 12))) {   //12 is the length of two ETH addresses; after the src and dst is the type field
	case ETH_P_ARP:
		DEBUG(10, printk("klunar - arp detected\n"));

                //The following lines are added by me***********************************************************
                //arp = skb->nh.arph;
                //dst = *((struct in_addr *)arp->ar_tip);
                dst = *(struct in_addr *)(skb->data + 38);       //target address of ARP request; hinted by the 1st param in lunar_do_arp
		DEBUG(10, printk(" klunar.c: the target IP is 0x%08x\n", dst.s_addr));
                struct neighbour *neigh = neigh_lookup(&arp_tbl, &(dst.s_addr), dev);
                if(neigh) {
		  length =(neigh->arp_queue).next->len;  //the "len" includes mac header????? 
		  //length = 1000;   //test;
                  DEBUG(10, printk("the data packet is %d long\n", length));
                  //double Tdata = Trts+Tcts+Tack+8*(length+sizeof(sapfsel_t))/dRate+4*DSSS_SIFSTime+2*DSSS_SlotTime;
		  Our_Proc_File->read_proc(buffer, NULL, 0,0, NULL, NULL);
		  char* pStr=buffer;
		  while(*pStr!='X') 
			pStr++;

                  *pStr = '\0';
                  
                  int rate = simple_strtol(buffer, NULL, 0);
		  DEBUG(10, printk(" the rate is %d\n", rate));
                  //rbw = (long)(rate*Tdata*dRate/1000000);     //the unit of BW is kbps; how to get rate from application????????
		  DEBUG(10, printk("Tack = %ld, sifs = %d, dRate = %d\n", Trts, DSSS_SIFSTime, dRate));
		  //rbw = (unsigned long)(rate*dRate*(Tack+8*1000000*(length+8)/dRate+2*DSSS_SIFSTime+2*DSSS_SlotTime)/1000);
		  unsigned long oh = (unsigned long)(rate*dRate*(Tack+2*DSSS_SIFSTime+2*DSSS_SlotTime)/1000);
		  DEBUG(10, printk("consumed bw due to mac oh is %ld\n", oh)); 
		  //rbw = (unsigned long)(rate*8*1000000*(length+8)+rate*dRate*(Tack+2*DSSS_SIFSTime+2*DSSS_SlotTime))/1000);
		  rbw = (unsigned long)(rate*8*1000*(length+8)+oh);
		  DEBUG(10, printk("expanded required bw is %ld\n", rbw)); 
		  rbw = (unsigned long)(rbw/1000000);
		  //rbw = (double)(20*Tdata*dRate/1000);         //for debuging, set rate = 20pkts/s
		  //rbw = 100.0;
		  DEBUG(10, printk(" the rate is %d and the required bw is %ld\n",rate, rbw));
                }
                
                else {
                  //printf("can't find the corresponding entry in the arp_tbl\n");
                  return -1;
                }
		//**********************************************************************************************
		
		lunar_do_arp(*(struct in_addr *)(skb->data + 38), rbw);  //add the second parameter rbw
                kfree_skb(skb);
		break;
		
	case ETH_P_IP:
		if (stats) {
			stats->tx_packets++;
			stats->tx_bytes += skb->len;
		}
	{
              
		int i;
		DEBUG(10, printk("IP data (%d):", skb->len));
		for (i = 0; i < 50; i++)
			DEBUG(10, printk(" %02x", skb->data[i]));
		DEBUG(10, printk("\n"));
	}
		pkt.data = skb->data + dev->hard_header_len;
		pkt.len = skb->len - dev->hard_header_len;
		pkt.ifnr = -1;

		if (pkt.len >= (20+8+44)  && *((char*)pkt.data+9) == 17 && // UDP
		    ntohs(*(short*)((char*)pkt.data+20)) == 68 &&  // DHCP clnt
		    ntohs(*(short*)((char*)pkt.data+22)) == 67)    // DHCP srvr
			lunar_do_dhcp(&pkt);
		else if (!memcmp(skb->data, "\xff\xff\xff\xff\xff\xff", ETH_ALEN)){ // broadcast
			if (bcast_selector) {
				DEBUG(10, printk("bcast_selector already exists\n"));
				sapf_demux(mysapfhandle, bcast_selector, &pkt);

			}
			else {
				DEBUG(10, printk("bcast_selector not existing\n"));
				lunar_do_bc_setup(&bcast_selector, &pkt);
			}
		} else { // unicast
		        //************following lines are added by me***********************
                        unsigned int nh_src = (skb->nh.iph)->saddr;   //assuming a IP network layer
                        unsigned int nh_dst = (skb->nh.iph)->daddr;
		        int ndx = add_flow(nh_src, nh_dst);
                        if (ndx < 0) {
		            kfree_skb(skb);
			    //printf("No such entry in flows table\n");
                            return 0;
                        }
                        if(flows[ndx].status != 0 &&flows[ndx].status != 3 && flows[ndx].status != 4) {
                          flows[ndx].status = 3;
                          flows[ndx].last_bw = flows[ndx].bw;
                        } else {
                          if(flows[ndx].status == 0) {
		            kfree_skb(skb);
			    //NETBOX_Instance netbox_arpdelete(nh_dst, lunardev);
			    //printf("Flow status (%d - %d) is %d, return!\n", nh_src,nh_dst,0);
                            return 0;
                          }
                        }
                        //******************************************************************
			sel = 0;
			memcpy(((char*)&sel) + sizeof(sel) - ETH_ALEN,
			       skb->data,ETH_ALEN);
			sapf_demux(mysapfhandle, sel, &pkt);  //data in pkt doesn't include the hard header
			
		}
		kfree_skb(skb);
		break;
	default:
		kfree_skb(skb);
		break;
	}

        return 0;
}


static int klunar_stop(struct net_device *dev)
{
	if (basedev) {
		int i = basedev->ifindex;
		basedev = 0;
		netbox_dropL2if(i);
	}
        netif_stop_queue(dev);

        return 0;
}


static struct net_device_stats *klunar_get_stats(struct net_device *dev)
{
        return (struct net_device_stats *) dev->priv;
}

int procfile_read(char *buf, char **buffer_location, off_t offset, int buffer_length, int *eof, void *data)
{
	        int ret;

                printk(KERN_INFO "profile_read(/proc/%s) called\n", PROCFS_NAME);

	        if(offset > 0) {
	                ret = 0;
	        } else {
	                memcpy(buf, procfs_buffer, procfs_buffer_size);
	                ret = procfs_buffer_size;
	        }
	        return ret;
}

int procfile_write(struct file *file, const char *buffer, unsigned long count, void *data)
{
	        procfs_buffer_size = count;
	        if(procfs_buffer_size > PROCFS_MAX_SIZE) {
	                procfs_buffer_size = PROCFS_MAX_SIZE;
		}

		if(copy_from_user(procfs_buffer, buffer, procfs_buffer_size)) {
		        return -EFAULT;
		}
		DEBUG(10, printk("procfs_buffer_size = %ld\n", procfs_buffer_size));
		return procfs_buffer_size;
}



// ----------------------------------------------------------------------

static int __init init_klunar(void)
{
	struct net_device *dev = 0;

	DEBUG(1, printk("klunar init module\n"));

	parse_mackill();

        lunardev = init_etherdev(0, sizeof(struct net_device_stats));
        if (!lunardev) {
		DEBUG(10, printk("klunar_init: no lunardev\n"));
		sapf_close(mysapfhandle);
                return -ENOMEM;
	}
	SET_MODULE_OWNER(lunardev);
	DEBUG(1, printk("klunar - lunar device is %s\n", lunardev->name));

	if (basedevname) {
		dev = dev_get_by_name(basedevname);
		if (!dev) {
			DEBUG(1, printk("klunar - device %s not found\n", basedevname));
		} else {
			dev_put(dev);
			if (dev == lunardev) {
				DEBUG(1, printk("klunar - invalid base device %s\n",
				       basedevname));
				dev = 0;
			}
		}
		basedev = dev;
		if (dev)
			netbox_addL2if(dev->ifindex, mysapfhandle);
	} else {
		read_lock(&dev_base_lock);
		for (dev = dev_base; dev != NULL; dev = dev->next) {
			if (dev->get_wireless_stats)
				break;
		}
		if (dev) {
			basedev = dev;
			DEBUG(10, printk("klunar - wireless interface is %s\n",
				 dev->name));
			netbox_addL2if(dev->ifindex, mysapfhandle);
		}
		read_unlock(&dev_base_lock);
	}

        lunardev->open = &klunar_open;
        lunardev->hard_start_xmit = &klunar_start_xmit;
        lunardev->stop = &klunar_stop;
        lunardev->get_stats = &klunar_get_stats;

	lunardev->mtu -= sizeof(sapfsel_t);
        lunardev->tx_queue_len = 3;
        *(short*)(lunardev->dev_addr) = htons(0x00ff);
        get_random_bytes(lunardev->dev_addr + 2, 4);

	if (lunar_init()) {
		unregister_netdev(lunardev);
		return -1;
	}

        register_netdevice_notifier(&klunar_notifier);
        proc_net_create("lunar", 0, lunar_proc_info);

	//*************create a /proc entry**************
	Our_Proc_File = create_proc_entry(PROCFS_NAME, 0644, NULL);

	if(Our_Proc_File == NULL) {
		remove_proc_entry(PROCFS_NAME, &proc_root);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n", PROCFS_NAME);
		return -ENOMEM;
	}

	Our_Proc_File->read_proc = procfile_read;
	Our_Proc_File->write_proc = procfile_write;
	Our_Proc_File->owner = THIS_MODULE;
	Our_Proc_File->mode = S_IFREG|S_IRUGO;
	Our_Proc_File->uid = 0;
	Our_Proc_File->gid = 0;
	Our_Proc_File->size = 37;

	printk(KERN_INFO "/proc/%s created\n", PROCFS_NAME);
	
	//***************************************************

        //I assume this function if what initiates the lunar module (to be veified)
        //when initiated, start learning the bandwidth information
        //init_check_bw();

	return 0;
}

static void __exit exit_klunar(void)
{
	DEBUG(1, printk("klunar unloading module\n"));

        proc_net_remove("lunar");
        unregister_netdevice_notifier(&klunar_notifier);
	unregister_netdev(lunardev);
    
	//***************remove /proc entry************
	remove_proc_entry(PROCFS_NAME, &proc_root);
	printk(KERN_INFO "/proc/%s removed\n", PROCFS_NAME);
	lunar_exit();
}


module_init(init_klunar);
module_exit(exit_klunar);

/*int procfile_read(char *buffer, char **buffer_location, off_t offset, int buffer_length, int *eof, void *data)
{
        int ret;

	printk(KERN_INFO "profile_read(/proc/%s) called\n", PROCFS_NAME);

	if(offset > 0) {
		ret = 0;
	} else {
		memcpy(buffer, procfs_buffer, procfs_buffer_size);
		ret = procfs_buffer_size;
	}
	return ret;
}
 
int procfile_write(struct file *file, const char *buffer, unsigned long count, void *data)
{
	procfs_buffer_size = count;
	if(procfs_buffer_size > PROCFS_MAX_SIZE) {
		procfs_buffer_size = PROCFS_MAX_SIZE;
	}

	if(copy_from_user(procfs_buffer, buffer, procfs_buffer_size)) {
		return -EFAULT;
	}
	return procfs_buffer_size;
}*/

// eof
