/* knetbox.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   an internal API for access to layer 2+3 service points
 *
 * Copyright (C) 2001-2004 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *   2004-01-09 added mackill filter in l2recv()
 *   2003-11-24 added getHID128()
 *   2003-11-20 added netbox_hasDGF(), removed netbox_arpannounce()
 */

#include "../include/netbox.h"
#include "../include/netbox_decl.h"
//#include "linux/modversions.h"

//#include "../include/sapf.h"
// ----------------------------------------------------------------------

//static struct sapfpkt_s packet_type;
static struct packet_type sapfpacket;
static int myhandle;
static char hid[NETBOX_HID_LEN];


char* netbox_eth2str(char *data)
{
	static char s[30];

	sprintf(s, "%02x:%02x:%02x:%02x:%02x:%02x",
		 (unsigned char) data[0], (unsigned char) data[1],
		 (unsigned char) data[2], (unsigned char) data[3],
		 (unsigned char) data[4], (unsigned char) data[5]);

	return s;
}

// ----------------------------------------------------------------------

static netbox_timer_t *timerlist;
static struct timer_list netbox_timer;

static void timer_click(unsigned long dummy)
{
	if (timerlist) {
		netbox_timer_t *t = timerlist;
		timerlist = timerlist->next;
		if (timerlist) {
			netbox_timer.expires = timerlist->expires;
			add_timer(&netbox_timer);
		}
		t->fct(t->data);
	}
}


void netbox_newtimer(netbox_timer_t *t,
		     void (*fct)(void *), void *data, unsigned long delta_usec)
{
	netbox_timer_t **pp;
	unsigned long expires = jiffies + HZ * delta_usec / 1000000;

	DEBUG(10, printk("newtimer %ld\n", delta_usec));
	if (!timerlist)
		pp = &timerlist;
	else {
		for (pp = &timerlist; ; pp = &((*pp)->next)) {
			if (!*pp || time_after((*pp)->expires, expires))
				break;
		}
	}
	t->next = *pp;
	t->expires = expires;
	t->fct = fct;
	t->data = data;
	*pp = t;

	if (t == timerlist) {
		if (t->next)
			del_timer(&netbox_timer);
		netbox_timer.expires = timerlist->expires;
		add_timer(&netbox_timer);
	}

	return;
}

void netbox_deltimer(netbox_timer_t *t)
{
	netbox_timer_t **pp;

	if (t == timerlist) {
		del_timer(&netbox_timer);
		timerlist = timerlist->next;
		if (timerlist) {
			netbox_timer.expires = timerlist->expires;
			add_timer(&netbox_timer);
		}
		return;
	}

	for (pp = &timerlist; *pp; pp = &((*pp)->next)) {
		if (*pp == t) {
			*pp = t->next;
			return;
		}
	}
}

// ----------------------------------------------------------------------

int netbox_init()
{
        init_timer(&netbox_timer);
	netbox_timer.function = timer_click;

	get_random_bytes(hid, sizeof(hid));

	return 0;
}


void netbox_fini()
{
	if (timerlist)
		del_timer(&netbox_timer);
}

// ----------------------------------------------------------------------

static int l2recv(struct sk_buff *skb,
		  struct net_device *indev,
		  struct packet_type *pt)
{
  //This function is an analog to selnet::macinput()******************
  //do the similar modification like in macinput**********************
	struct sapfpkt_s *p = (struct sapfpkt_s *) &skb->cb;
	sapfsel_t dest;
	int i;
	uint32_t src, dst;
	DEBUG(10, printk("knetbox L2recv\n"));
        //*************************************************************
	//here, we need to get the source and destination IP 
	//char *ipl;     //IP layer
	//ipl = skb->nh.raw + sizeof(sapfsel_t);   //skip the selector field
	//memcpy(&src, ipl+12, sizeof(uint32_t));
	//memcpy(&dst, ipl+16, sizeof(uint32_t));
        //unsigned int src = (skb->nh.iph).saddr;
        //unsigned int dst = (skb->nh.iph).daddr;
        //*************************************************************
	for (i =0; i < mackill_cnt; i++) {
		if (!memcmp(skb->mac.ethernet->h_source,
			    mackill_list[i], ETH_ALEN)) {
			DEBUG(10, printk("knetbox l2recv: dropping packet from %s\n", netbox_eth2str(skb->mac.ethernet->h_source)));
			goto done;
		}
	}
        
        //*************************************************************
	//Here we need to get the src and dst IP
	uint8_t *ver;  //1-byte version
	uint8_t version;
	ver = skb->data + sizeof(sapfsel_t);    //skip the selector field
	memcpy(&version, ver, sizeof(char));
	DEBUG(10, printk("The version is %d\n", version));
	//memcpy(&dst, skb->data + sizeof(sapfsel_t) + 16, sizeof(uint32_t));
	//DEBUG(10, printk("the dst is 0x%8x; bcast is 0x%8x\n", dst, htonl(LUNAR_IP_BCAST)));

	if(version != 0 && version != 1 ) {
            memcpy(&src, skb->data + sizeof(sapfsel_t) + 12, sizeof(uint32_t));
            memcpy(&dst, skb->data + sizeof(sapfsel_t) + 16, sizeof(uint32_t));
	    DEBUG(10, printk("the dst is 0x%8x; bcast is 0x%8x\n", dst, htonl(LUNAR_IP_BCAST)));
	    DEBUG(10, printk("Got Source and Dest IP addresses\n"));
	    if(dst != htonl(LUNAR_IP_BCAST)) {
                 int ndx = add_flow(src,dst);
                 if(ndx < 0) {
		    DEBUG(10, printk(" L2recv -- no such entry in flow table\n"));
		    p = NULL;
		    kfree_skb(skb);
	            //printf("macinput -- no such entry in flows\n");
                    return -1;
                 }      
                 if(flows[ndx].status != 0 && flows[ndx].status !=3 &&flows[ndx].status != 4) {
                    //fprintf(stderr, "node %d recv a pkt sent from %d to %d-- change status to 3\n",My_IP, src_,dst_);
                    flows[ndx].status = 3;
                    flows[ndx].last_bw = flows[ndx].bw;
                    //flows[ndx].data = 0;
                 } else {
                    //if (flows[ndx].status < 3) {
                    if (flows[ndx].status == 0) {
			p = NULL;
			kfree_skb(skb);
                       //fprintf(stderr, "node %d recv a pkt sent from %d to %d, but the flow status is %d\n",My_IP,src_,dst_,flows[ndx].status);
                       return -1;
                    }
                 }
                 //fprintf(stderr, "node %d recv a packet from %d to %d\n",My_IP,src_,dst_);
        
                 NETBOX_Instance netbox_deltimer(&(flows[ndx].timer));
                 NETBOX_Instance netbox_newtimer(&(flows[ndx].timer), _VOID_VOID_ data_tout, flows+ndx, DATA_TIMEOUT);
	     }
	    }
        //*************************************************************************************************
        
	memset(p, 0, sizeof(*p));
	p->snr = 0;
	p->ifnr = indev->ifindex;
	p->data = skb->nh.raw + sizeof(sapfsel_t);    //skb->nh.raw includes the selector field. this can be seen from netbox_l2send
	p->len = skb->len - sizeof(sapfsel_t);

	memcpy(&dest, skb->nh.raw, sizeof(dest));

	DEBUG(10, printk("sapf packet, %d (%d) bytes via %s: sel=0x%016Lx, refcnt=%d\n",
		skb->len, skb->len - sizeof(sapfsel_t),
		indev ? indev->name : "no-dev", ntohsel(dest),
		indev->refcnt.counter));

	sapf_demux(*(int*)pt->data, dest, p);        //p only includes upper-layer data
        //sapf_demux(*(int*)pt->af_packet_priv, dest, p); //different kernel version

 done:
	kfree_skb(skb);
	return 0;
}


int netbox_addL2if(int ifnr, int handle)
{
	struct net_device *dev = dev_get_by_index(ifnr);

	DEBUG(1, printk("netbox addL2if ifnr=%d\n", ifnr));

	if (sapfpacket.func) {
		DEBUG(1, printk("  sapf already bound\n"));
		return -1;
	}
	if (dev) {
		DEBUG(1, printk("  dev=%s\n", dev->name));
	} else {
		DEBUG(1, printk("  no dev!!\n"));
	}

	myhandle = handle;

	//	MOD_INC_USE_COUNT;

	sapfpacket.type = htons(SAPF_ETHERTYPE);
	sapfpacket.dev = dev;
	sapfpacket.func = l2recv;
	sapfpacket.data = &myhandle; //*******address********
        //sapfpacket.af_packet_priv=&myhandle; //due to different kernel
	dev_add_pack(&sapfpacket);

	return 0;
}


int netbox_dropL2if(int ifnr)
{
	if (!sapfpacket.func)
		return -1;

	DEBUG(10, printk("netbox dropL2if\n"));

	dev_remove_pack(&sapfpacket);
	neigh_ifdown(&arp_tbl, sapfpacket.dev);

	dev_put(sapfpacket.dev);
	sapfpacket.func = 0;

	return 0;
}


int netbox_getHID128(unsigned char *buf)
{
	memcpy(buf, hid, sizeof(hid));
	return 0; // success
}


int netbox_getL2addr(int ifnr, struct sockaddr *sa)
{
	struct net_device *dev = dev_get_by_index(ifnr);
	int i;

	if (!dev) {
		DEBUG(0, printk("knetbox getL2addr: wrong index %d\n", ifnr));
		return -1;
	}

	memcpy(sa->sa_data, dev->dev_addr, ETH_ALEN);

	DEBUG(10, printk("getL2addr:"));
	for (i = 0; i < ETH_ALEN; i++)
		DEBUG(10, printk("%02x ", (unsigned char) sa->sa_data[i]));
	DEBUG(10, printk("\n"));

	dev_put(dev);

	return 0;
}


int netbox_getL3addr(struct in_addr *name, struct net_device *dev)
{
	struct in_device *indev = (struct in_device*) dev->ip_ptr;

	if (!indev)
		return -1;
	if (!indev->ifa_list)
		return -1;
	name->s_addr = indev->ifa_list->ifa_local;
	DEBUG(10, printk("local %08x\n", ntohl(name->s_addr)));

	return 0;
}


// IP addr in network order
int netbox_arpset(struct in_addr addr, struct sockaddr *hw_addr, struct net_device *dev)
{
	struct neighbour *neigh;
	int i;

	DEBUG(10, printk("arpset: %08x ", ntohl(addr.s_addr)));
	for (i = 0; i < ETH_ALEN; i++) {
		DEBUG(10, printk(" %02x", (unsigned char) hw_addr->sa_data[i]));
	}
	DEBUG(10, printk("\n"));

	neigh = __neigh_lookup_errno(&arp_tbl, &(addr.s_addr), dev);
	//        err = PTR_ERR(neigh);
        if (!IS_ERR(neigh)) {
		DEBUG(10, printk("neighb update\n"));
		neigh->parms->delay_probe_time = 0;
                neigh_update(neigh, hw_addr->sa_data, NUD_REACHABLE, 1, 0);
                neigh_release(neigh);
        }

	return 0;
}


// IP addr in network order
int netbox_arpdelete(struct in_addr addr, struct net_device *dev)
{
	struct neighbour *neigh;

	DEBUG(10, printk("arpdelete %08x\n", ntohl(addr.s_addr)));

	neigh = neigh_lookup(&arp_tbl, &(addr.s_addr), dev);
        if (neigh) {
                neigh_update(neigh, 0, NUD_NONE, 1, 0);
                neigh_release(neigh);
        }

	return 0;
}


int netbox_l2send(int ifnr, struct sockaddr *dest, sapfsel_t sel,
			 void *data, int len)
     // selector must be in network order
{
	struct net_device *dev;
	struct sk_buff *skb;
	char *cp;
	static struct sockaddr broadcast = {AF_UNSPEC, {0xff, 0xff, 0xff, 0xff, 0xff, 0xff}};

	if (!dest)
		dest = &broadcast;

	DEBUG(10, printk("netbox_l2send %s 0x%016Lx (%d bytes)\n",
	       netbox_eth2str(dest->sa_data), ntohsel(sel), len));

	dev = dev_get_by_index(ifnr);
	if (!dev) {
		DEBUG(1, printk("netbox_l2send: wrong ifnr\n"));
		return 0;
	}
	DEBUG(10, printk("  outgoing dev is %s\n", dev->name));

	skb = alloc_skb(dev->hard_header_len + 15 +
			sizeof(sel) + len, GFP_ATOMIC);
	if (!skb) {
		DEBUG(1, printk("  alloc_skb failed\n"));
		return 0;
	}

	skb_reserve(skb, (dev->hard_header_len+15)&~15);
	skb->nh.raw = skb->data;
	skb->protocol = htons(SAPF_ETHERTYPE);
	skb->dev = dev;

	cp = skb_put(skb, sizeof(sel) + len);
	memcpy(cp, &sel, sizeof(sel));
	memcpy(cp + sizeof(sel), data, len);

	if (dev->hard_header) {
		dev->hard_header(skb, dev, SAPF_ETHERTYPE,
				 dest->sa_data, 0, skb->len);
		dev_queue_xmit(skb);
	} else {
		DEBUG(1, printk("netbox_l2send: no hard header, what to do?\n"));
	}

	dev_put(dev);
	return len;
}


int netbox_l3deliver(int ifnr, void *data, int len)
{
        struct net_device *dev = dev_get_by_index(ifnr);
        struct sk_buff *skb;

        if (!dev)
                return 0;

	DEBUG(10, printk("netbox_l3deliver %d Bytes\n", len));


	skb = alloc_skb(len + 14, GFP_ATOMIC);
	if (!skb) {
		dev_put(dev);
		return 0;
	}
	skb->dev = dev;

	skb->mac.raw = skb->data;
	skb_reserve(skb, 14);
	memcpy(skb_put(skb, len), data, len);
	memcpy(skb->mac.raw, dev->dev_addr, ETH_ALEN);
	memset(skb->mac.raw + ETH_ALEN, 0, ETH_ALEN);
	*((short*)(skb->mac.raw + 2 * ETH_ALEN)) = htons(ETH_P_IP);
	skb->protocol = htons(ETH_P_IP);

	netif_rx(skb);

	dev_put(dev);
	return 0;
}


/* int netbox_hasDGW(void);

   returns true when the host has a default GW entry in its
   routing table and should forward packets to the fixed
   internet.
*/
int netbox_hasDGW(void)
{
/*
	struct nlmsghdr nl;
	struct rtmsg rtm;
	struct kern_rta rta;
	struct rtentry r;
	int rc;
	struct fib_table *tb;

	memset(&r, 0, sizeof(r));
	r.rt_dst = ;
	r.rt_flags = ;
	r.rt_genmask = ;
	r.rt_gateway = ;
	rc = fib_convert_rtentry(SIOCDELRT, &nl, &rtm, &rta, &r);
	tb = fib_get_table(rtm.rtm_table);
*/
	return 0;
}

void data_tout(void* vp)
{
  struct flow_s *n = (struct flow_s *)vp;
  if (n->status == 3) {
    n->status = 0;
  }
  //printf("node data_tout: status of flow(%d - %d) changed to 0\n", n->src, n->dst);
}

// eof
