/* selnet.cc
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   Lunar port to ns-2 
 *
 * Copyright (C) 2001-2003 <christian.tschudin@unibas.ch>
 * Ported by Evgeny Osipov <evgeny.osipov@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */
#ifndef NS_PORT
#define NS_PORT
#endif


#include "../include/platform.h"

#include <errno.h>



#include "delay.h"
#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "rtp.h"
#include "tcp.h"
#include "mac-802_11.h"
#include "selnet.h"
#include "sapf_ns.h"
#include "netbox_ns.h"
#include "topography.h"
#include "cmu-trace.h"
#include "mobilenode.h"
#include "tools/random.h"

#include <address.h>

static class SELNETClass : public TclClass {
public:
        SELNETClass() : TclClass("SELNET") {}
        TclObject* create(int, const char*const* argv) {
                return (new SELNET(argv[4], argv[5], argv[6], argv[7], argv[8],argv[9]));
        }
} class_selnet;

int hdr_selnet::offset_;

static class SELNETHeaderClass : public PacketHeaderClass {
public:
        SELNETHeaderClass() : PacketHeaderClass("PacketHeader/SELNET",
                                            sizeof(hdr_selnet)) { 
		bind_offset(&hdr_selnet::offset_);
	}
        //void export_offsets() {
        //        field_offset("sel", OFFSET(hdr_selnet,sel));
        //}
} class_hdr_selnet;

int hdr_xrp::offset_;

static class XRPHeaderClass : public PacketHeaderClass {
public:
        XRPHeaderClass() : PacketHeaderClass("PacketHeader/XRP",
                                            sizeof(hdr_xrp)) { 
		bind_offset(&hdr_xrp::offset_);
	}
        //void export_offsets() {
        //        field_offset("data", OFFSET(hdr_xrp,data)); 
        //}
} class_hdr_xrp;




unsigned int Random_byte=155;
#include "../src/lunar.c"
                                                                                
#include "../src/xrp.c"
                                                                                
#include "netbox_ns.cc"
                                                                                


SELNET::SELNET(const char *tclnode, const char *tclmac, const char *tclll,const char *arptable,const char *ifq, const char *phy) : LinkDelay()  
{ 
  //fprintf(stderr, "Here is SELNET::SELNET");
  node_ = (MobileNode*) TclObject::lookup(tclnode);
  assert(node_);
  My_IP=node_->address();
  Node *tmp=(Node*)node_;

  mac_ = (Mac802_11*) TclObject::lookup(tclmac);
  assert(mac_);
  
  ll_ = (LL*) TclObject::lookup(tclll);
  assert(ll_);
  
  arptable_ = (ARPTable*) TclObject::lookup(arptable);
  assert(arptable_);
  
  ifq_ = (NsObject*) TclObject::lookup(ifq);
  assert(arptable_);
  //fprintf(stderr, "We want to figure out the PHY\n");

  phy_ = (WirelessPhy*) TclObject::lookup(phy);
  assert(phy_);

  mac_->set_phy(phy_);
 
  //bw_timer = new NETBOXTimer();
  //cs_timer = new NETBOXTimer();
  bw_timer.set_callback(this, debuglevel);
  cs_timer.set_callback(this, debuglevel);
  //fprintf(stderr, "Got each component\n");
  // The following lines were added by me
  local_bw_cs= Bmax;
  csn_bw = Bmax;
  csn_btime = 0;
  //bw_cs_data = Bmax;
  Trts = mac_->txtime_rts();
  Tcts = mac_->txtime_cts();
  Tack = mac_->txtime_ack();
  mac_datarate = mac_->mac_data_rate();
  //fprintf(stderr, "Begin to check available bandwidth\n");
  //init_check_bw();
  // ************************************  
	
  /*INITIALIZE MY FIELDS FIRST*/
  // XRP_CTYPE_UNKNOWN:
  xrp_param_len[0]=  0;
  // XRP_CTYPE_SEL:
  xrp_param_len[1]=   sizeof(struct xrp_common_s) + sizeof(sapfsel_t);
  // XRP_CTYPE_IPv4
  xrp_param_len[2]= sizeof(struct xrp_common_s) + sizeof(uint32_t);
  // XRP_CTYPE_IPv6
  xrp_param_len[3]= sizeof(struct xrp_common_s) + 16;
  // XRP_CTYPE_SEL_ETH
  xrp_param_len[4]= sizeof(struct xrp_common_s) + sizeof(sapfsel_t) + ETH_ALEN;
  // XRP_CTYPE_SEL_UDP
  xrp_param_len[5]= sizeof(struct xrp_common_s) + sizeof(sapfsel_t) + sizeof(uint32_t)
    + sizeof(short);
  xrp_param_len[6]= sizeof(struct xrp_common_s) + sizeof(double);
  //fprintf(stderr, "xrp_param_len[6] is %d\n",xrp_param_len[6]);
  // XRP_CTYPE_HID128
  xrp_param_len[XRP_PARTYPE_COUNT]=  sizeof(struct xrp_head_s) + 16;
  
  
  stats=new struct net_device_stats;
  /*create sapf object here*/
  sapf_=new SAPF(this);
  sapf_->sapf_init();
  //Initialize FIB array
  for(int i=0;i<SAPF_MAX_FIB_TABLES;i++)
    {
      sapf_->fib[i]=new struct fibtable_s; 
      memset(sapf_->fib[i], 0, sizeof(struct fibtable_s));
    }
  
  // the following two lines are added by me
  //bw_timer = new NETBOXTimer;
  //cs_timer = new NETBOXTimer;


  bind("conf_min_ring", &conf_min_ring);
  bind("conf_max_ring", &conf_max_ring);
  bind("conf_timeout_per_ring", &conf_timeout_per_ring);
  bind("conf_retry_wait", &conf_retry_wait);
  bind("conf_max_retries", &conf_max_retries);
  bind("conf_fri", &conf_fri);
  bind("conf_fri_jitter", &conf_fri_jitter);
  bind("debuglevel", &debuglevel);
  // the following lines were added by me
  bind("CHECK_TIME", &CHECK_TIME);
  //fprintf(stderr, "CHECK_TIME is %4f\n", CHECK_TIME);
  bind("CHECK_PERIOD",&CHECK_PERIOD);
  //fprintf(stderr, "CHECK_PERIOD is %4f\n",CHECK_PERIOD);
  bind("DATA_TIMEOUT", &DATA_TIMEOUT);
  bind("RESV_TIMEOUT", &RESV_TIMEOUT);
  
  forwarding="yes";
  //fprintf(stderr, "Begin to check available bandwidth\n");
  init_check_bw();
  
  /*create netbox object here*/
  netbox_=new NETBOX(this);
  for(int i=0;i<LUNAR_MAX_PEERS;i++)
    { 
      peers[i].timer.set_callback(this, debuglevel);
    }
  
  lunardev=new struct net_device;
  
  lunardev->ifindex=0;
  lunar_init(); 
  
  DEBUG(10, printk("MYIP=%d\n",My_IP));
  DEBUG(10, printk("initialized \n")); 
  //fprintf(stderr, "Here is SELNET::SELNET finished");
}

int
SELNET::command(int argc, const char*const* argv)
{
	if (argc == 2 && strcasecmp(argv[1], "reset") == 0) {
	}
	return LinkDelay::command(argc, argv);
}

void SELNET::handle(Event* e)
{
  recv((Packet *)e,0);
}
//#include "../src/lunar.c"

//#include "../src/xrp.c"

//#include "netbox_ns.cc"

void SELNET::init_check_bw()
{
  //fprintf(stderr, "Here is SELNET::init_check_bw\n");
  //fprintf(stderr, "bw_timer is 0x%x\n",&bw_timer);
  //fprintf(stderr, "cs_timer is 0x%x\n",&cs_timer);  
  check_bandwidth_cs((void*) 0);
}

void SELNET::recv(Packet* p, Handler* /*h*/)
{	
  //fprintf(stderr, "Here is SELNET::recv");
  hdr_cmn *ch = HDR_CMN(p);
  DEBUG(10, printk("SELNET::recv:PACKET - Direction= %d (DOwN=%d UP=%d default=-1)\n",ch->direction(),hdr_cmn::DOWN,hdr_cmn::UP));

  // If direction = UP, then pass it up the stack
  // Otherwise, set direction to DOWN and pass it down the stack
  if(ch->direction() == hdr_cmn::UP) {
    if(ch->ptype_ == PT_SELNET || ch->ptype_ == PT_XRP) {
         //hdr_selnet *sh = HDR_SELNET(p);
         //fprintf(stderr, "SELNET::recv--selector in pkt header is 0x%016Lx\n",sh->selector());
         macinput(p);}
    else //end Lunars'extensions //this should never happen in Lunar
      // ALL PACKETS ARE SELNET
      ll_ ? ll_->recv(p,0) : drop(p); // send to ll it will handle it.
    return;
  }
  
  ch->direction() = hdr_cmn::DOWN;
  ipinput(p);
}


 
int
SELNET::ipinput(Packet *p)
{
  //fprintf(stderr, "Here is SELNET::ipinput");
  /*THIS METHOD IS AN ANALOGUE TO klunar_start_xmit()*/
  Scheduler& s = Scheduler::instance();

  hdr_cmn *ch = HDR_CMN(p);
  char *mh=(char*)HDR_MAC(p);
  hdr_ll	*lh = HDR_LL(p);
  hdr_arp	*ah = HDR_ARP(p);
  hdr_selnet *sh = HDR_SELNET(p);
  hdr_xrp *xh = HDR_XRP(p);
  hdr_ip *ih = HDR_IP(p);

  struct hdr_mac802_11 *mh_802_11=(struct hdr_mac802_11 *)mh;
  
  nsaddr_t dst_= (nsaddr_t)Address::instance().get_nodeaddr(ih->daddr());
  
  nsaddr_t src_ = (nsaddr_t)Address::instance().get_nodeaddr(ih->saddr());
   //make real code happy
  struct in_addr addr;
  struct net_device *dev=basedev;

  int len = ch->size();
  int length = arptable_->incoming_pkt_size_;
  //fprintf(stderr, "data packet length is %d\n", length);
  u_int16_t typeMAC= GET2BYTE(mh_802_11->dh_body);//this macro is from marshal.h included from mac-802_11.h

  /*THIS IS FROM klunar_start_xmit*/
  sapfsel_t sel;
  //sapfpkt pkt;// = new sapfpkt;
  nsaddr_t targetIP;
  int res;
  double Tdata, rbw;
  DEBUG(10, printk("klunar start_xmit (%d bytes)\n", len));

  switch (typeMAC) 
    {
    case ETHERTYPE_ARP:
      DEBUG(10, printk("klunar - arp detected\n"));
      targetIP=ah->arp_tpa;
      addr.s_addr=targetIP;
      //int length = arptable_->incoming_pkt_size_; 
      Tdata = Trts + Tcts + Tack +8*(length+sizeof(sapfsel_t)+sizeof(hdr_mac802_11))/mac_datarate + 4*DSSS_SIFSTime + 2*DSSS_SlotTime;
      //fprintf(stderr, "Tdata is %f\n", Tdata);
      rbw = ch->prate()*Tdata*mac_datarate/1000; // The unit of BW is kbps
      //fprintf(stderr, "SELNET: requested bw is %4f\n",rbw);
      res=lunar_do_arp(addr, rbw);
      if(res==-1)
	{
	  // DEBUG(10, printk("\n\n\n arp DROPED\n\n\n\n"));
	  //q_handler->handle((Event*)0);//keep the queue active
	};
      kfree_skb(p);//new packet will be created
      break;
    case ETHERTYPE_IP:
      //fprintf(stderr, "Here is SELNET::ipinput--recv a data packet from ip\n");
      if (stats) {
	stats->tx_packets++;
	stats->tx_bytes += len;
      }
      {
	int i;
	DEBUG(10, printk("IP data (%d):", len));
	DEBUG(10, printk("\n"));
      }
      *(xh->data)=(char)-1;//this is to indicate that packet is originated at this node
      pkt_ip.data = (char*)p;// put a pointer to real packet here
      pkt_ip.len = -1;//this is to indicate that there is a packet inside //skb->len - dev->hard_header_len;
      pkt_ip.ifnr = -1;
      int MAC_DA_INT;
      
      if (ETHER_ADDR(mh_802_11->dh_da)==MAC_BROADCAST)
	{
	  // broadcast
	  if (bcast_selector)
	    sapf_->sapf_demux(mysapfhandle, bcast_selector, &pkt_ip);
	  else
	    lunar_do_bc_setup(&bcast_selector, &pkt_ip);
	}     
      else { // unicast
        //fprintf(stderr, " SELNET send down data packet\n");
        int ndx = add_flow(src_, dst_);
        if (ndx < 0) {
            fprintf(stderr, " no such entry in flows\n");
            return 0;
        }
        //if (flows[ndx].status == 2 || flows[ndx].status == 5) {
        if(flows[ndx].status != 0 && flows[ndx].status !=3 &&flows[ndx].status != 4) {
            flows[ndx].status = 3;
            flows[ndx].last_bw = flows[ndx].bw;
            fprintf(stderr, "node %d Recv data, flow status change to 3 \n",My_IP);
        } else {
            //if (flows[ndx].status < 3) {
            if (flows[ndx].status == 0) {
                fprintf(stderr, "node %d Flow status(%d -> %d) is %d, return !\n", My_IP,flows[ndx].src,flows[ndx].dst,flows[ndx].status);
                return 0;
            }
        }
        //fprintf(stderr, "node %d recv a packet from IP layer to node %d\n",My_IP,dst_);
	sel = 0;
	MAC_DA_INT=htonl(GET4BYTE(mh_802_11->dh_da));
	memcpy(((char*)&sel) + sizeof(sel) - ETH_ALEN,
	       ((char*)&MAC_DA_INT),ETH_ALEN);
	sel|=htonsel(SELCLASS_LOCAL);
        
	sapf_->sapf_demux(mysapfhandle, sel, &pkt_ip);
	}
      break;
    default:
      kfree_skb(p);
      break;
    }
  
  return 0;
  
  

} 

void
SELNET::macinput(Packet *p)
{
  //this function is an analog to l2recv()

   Scheduler& s = Scheduler::instance();
   double now=s.clock();
  
  hdr_cmn *ch = HDR_CMN(p);
  hdr_ip *ih = HDR_IP(p);
  hdr_selnet *sh = HDR_SELNET(p);
  hdr_xrp *xh = HDR_XRP(p);
  hdr_tcp *th = HDR_TCP(p);
  sapfsel_t sel= sh->selector();
  
  nsaddr_t dst_= (nsaddr_t)Address::instance().get_nodeaddr(ih->daddr());
                                                                                
  nsaddr_t src_ = (nsaddr_t)Address::instance().get_nodeaddr(ih->saddr());

  //fprintf(stderr, " SELNET::macinput --the selector is 0x%16Lx\n", sel);
  
  //sapfpkt *pkt=new sapfpkt;
  if(ch->ptype()==PT_XRP)
    {
      //fprintf(stderr, "SELNET::macinput--received packet is routing pkt\n");
      memcpy(xrpdata, xh->data,2000);
	//xrpdata=xh->data;
      pkt.data = (char*)xrpdata;

      struct xrp_head_s *x = (struct xrp_head_s*) pkt.data;
      pkt.len = ch->size()-sizeof(sapfsel_t); //subtract the selector's size
      pkt.ifnr = -1;
      kfree_skb(p);
    } else //this is selnet packet but not RREQ/RREP
    {
       
        int ndx = add_flow(src_, dst_);
        if (ndx < 0) {
            fprintf(stderr, "macinput -- no such entry in flows\n");
            return;
        }
        
        //if (flows[ndx].status == 5 ||flows[ndx].status == 2) {
        if(flows[ndx].status != 0 && flows[ndx].status !=3 &&flows[ndx].status != 4) {
               fprintf(stderr, "node %d recv a pkt sent from %d to %d-- change status to 3\n",My_IP, src_,dst_);
               flows[ndx].status = 3;
               flows[ndx].last_bw = flows[ndx].bw;
               //flows[ndx].data = 0;
        } else {
           //if (flows[ndx].status < 3) {
           if (flows[ndx].status == 0) {
               fprintf(stderr, "node %d recv a pkt sent from %d to %d, but the flow status is %d\n",My_IP,src_,dst_,flows[ndx].status);
               return;
           }
        }
        //fprintf(stderr, "node %d recv a packet from %d to %d\n",My_IP,src_,dst_);
        
        NETBOX_Instance netbox_deltimer(&(flows[ndx].timer));
        NETBOX_Instance netbox_newtimer(&(flows[ndx].timer), _VOID_VOID_ data_tout, flows+ndx, DATA_TIMEOUT);
        
        ch->size()-=sizeof(sapfsel_t);
	pkt.data = (char*)p;
	pkt.len=-1;
	pkt.ifnr=-1;
    }

  DEBUG(10, printk("SELNET::macinput: Selector %016Lx\n", ntohsel(sel)));
  sapf_->sapf_demux(mysapfhandle,sel,&pkt);
 
}

void
SELNET::data_tout(void* vp)
{
     struct flow_s *n=(struct flow_s *)vp;
     if (n->status == 3 ) {
            n->status = 0;
     } 
     fprintf(stderr, "node data_tout: status of flow (%d - %d) changed to 0\n", n->src, n->dst);
}
             

#include "../src/lunar_init.c"


/*=======================================================================
SAPF staff: Constructor, TIMERÅ expire and redefined functtions
==========================================================================*/
//SAPF constructor
SAPF::SAPF(SELNET *a):sapf_gc_timer(this)
{
  selnet_=a;
  debuglevel=selnet_->debuglevel;
}

SAPFTimer::SAPFTimer(SAPF *a):TimerHandler()
{
  a_=a; 
  debuglevel=a_->debuglevel;
}

//NETBOX constructor
NETBOX::NETBOX(SELNET *a)
{
  selnet_=a;
  debuglevel=selnet_->debuglevel;
}
                                                    
                                                                           

void SAPF::add_timer(SAPFTimer *a)
{
  double timeout = a->expires;
  double now=Scheduler::instance().clock();
  a->resched(timeout); 
  a->expires=0.0;
}


void SAPF::del_timer(SAPFTimer *a)
{
  
  a->cancel();
 
}
void SAPF::get_random_bytes (sapfsel_t *sel, int nbytes)
{
  unsigned char *cp = (unsigned char *) sel;
  if (nbytes > 2)
    nbytes = 2;
  while (nbytes-- > 0) {
    *cp++ = Random::integer(256);
  }
}
void SAPFTimer::expire(Event *e)
{
  debuglevel=a_->debuglevel;
  double now=Scheduler::instance().clock();
  DEBUG(10, printk("sapf_gc:: now=%f\n", now));
  a_->sapf_gc(0);
}
void NETBOXTimer::expire(Event *e)
{
  double now=Scheduler::instance().clock();
  DEBUG(10, printk("\n\nNETBOX::timer:expire: %f \n", now));
  
  (a_->*fct)(data);
}
                                                                   
                                                                           


//void SELNET::init_check_bw() 
//{
//  check_bandwidth_cs((void*) 0);
//}
#include "../src/sapf.c"
 
  
 
 
 
 
