/* lunar.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   Lunar logic
 *
 * Copyright (C) 2001-2003 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */

//#include <random.h>
#include "../include/lunar.h"
#include "../include/lunar_decl.h"
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
//#include <stdlib.h>

// patch-me: we should pass "struct sockaddr" instead of a 32bit integer
// for the src parameter, so this cludge below would go away:
#ifndef NS_PORT
# define NO_IP_ADDR 0
#else
# define NO_IP_ADDR -1
#endif

// ----------------------------------------------------------------------

void NS_CLASS lunar_bcast_handler(sapfpkt *p, void *aux)
{
}

void NS_CLASS schedule_rreq(void *data, int len, unsigned long usec)
{
	int ifnr = basedev ? basedev->ifindex : 0;

	NETBOX_Instance netbox_l2send(ifnr, 0,
				      htonsel(SAPF_SELNET_LUNAR_RREQ),
				      data, len);
}

 int NS_CLASS peer_is_unknown(u_int32_t name)
{
	int i;

	for (i = 0; i < LUNAR_MAX_PEERS; i++)
		if (peers[i].used && peers[i].ia.s_addr == name)
			return 0;
	return 1;
}

 int NS_CLASS add_peer(struct in_addr name)
{
	int i, ndx = -1;

	for (i = 0; i < LUNAR_MAX_PEERS; i++) {
		if (ndx < 0 && !peers[i].used)
			ndx = i;
		else if (peers[i].ia.s_addr == name.s_addr)
			return i;
	}
	if (ndx >= 0) {

#ifndef NS_PORT
	  memset(peers + ndx, 0, sizeof(struct peer_s));
#endif
		peers[ndx].ia = name;
		peers[ndx].hops = 0;
	}
	return ndx;
}

int NS_CLASS add_flow(int src, int dst)
{
        int ndx = -1,i;
        //fprintf(stderr, "Here is add_flow\n");
        //fprintf(stderr, "add_flow: src is %d; dst is %d\n",src,dst);
        for (i = 0; i < LUNAR_MAX_FLOWS; i++) {
                if (ndx < 0 && flows[i].status == 0) {
                          //fprintf(stderr, "add_flow -- the empty entry is %d\n",i);
                          ndx = i;
                }
                else if (flows[i].src == src && flows[i].dst == dst && flows[i].status != 0 ) {
                          //fprintf(stderr, "the entry exists %d\n", i);
                          return i;
                }
        }
        if (ndx >= 0) {
                //fprintf(stderr, "entry not exist, but we have an empty one\n");
#ifndef NS_PORT 
                memset(flows + ndx, 0, sizeof(struct flow_s));
#endif
                flows[ndx].src = src;
                flows[ndx].dst = dst;
                flows[ndx].last_bw = 0;
                flows[ndx].bw = 0;
        } 
        return ndx;
}

// ----------------------------------------------------------------------

/*   mkSapfDest
 *   allocates an eth+selector (netorder) buffer, to be used as a
 *   network pointer in the SAPF module. The SAPF module will free
 *   this netptr when the SAPF entry is removed.
 */
struct sapfdest_s* NS_CLASS mkSapfDest(unsigned char *eth, sapfsel_t sel)
{
	struct sapfdest_s *dst =(struct sapfdest_s*)  kmalloc(sizeof(struct sapfdest_s), GFP_ATOMIC);
	unsigned char *cp;

	if (!dst)
		return 0;

	cp = eth ? eth : (unsigned char *) "\xff\xff\xff\xff\xff\xff";

	DEBUG(10, printk("mkSapfDest %02x:%02x:%02x:%02x:%02x:%02x 0x%016Lx\n",
		 cp[0], cp[1], cp[2], cp[3], cp[4], cp[5], ntohsel(sel)));

	dst->ifnr = basedev ? basedev->ifindex : 0;
	dst->sel = sel;
	memcpy(dst->dst.sa_data, cp, ETH_ALEN);

	return dst;
}

/*   mkNetPointer
 *   installs a forwarder entry, returns the new selector.
 */
sapfsel_t NS_CLASS mkNetPointer(unsigned char *eth, sapfsel_t sel)
{
	sapfsel_t ptr = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
        SAPF_Instance sapf_addCallback(mysapfhandle, ptr,NS_CLASS_addr L2sendto, mkSapfDest(eth, sel));
	return ptr;
}

sapfsel_t NS_CLASS mkHandlerPointer(sapfcallback_t fct, void *aux)
{
	sapfsel_t ptr =SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
        SAPF_Instance sapf_addCallback(mysapfhandle, ptr, fct, aux);
	return ptr;
}


// ----------------------------------------------------------------------

 void NS_CLASS arp_answer(int ndx)
{
	struct sockaddr sa;
	// sa.sa_family = ARPHRD_ETHER;
	// we use the low 48 bits, and all selectors are stored in nw order:
	memcpy(sa.sa_data,
	       ((char*) &(peers[ndx].fwdsel)) + sizeof(sapfsel_t) - ETH_ALEN, ETH_ALEN);
	NETBOX_Instance netbox_arpset(peers[ndx].ia, &sa, lunardev);
}

/* int NS_CLASS install_flow(int source, int dst, double bw)
{
        int ndx;
        ndx = add_flow(source, dst);
        if(ndx < 0) {
               DEBUG(10, printk("install flow: could not add flow\n"));
               return -1;
        }
        flows[ndx].bw = bw;
        NETBOX_Instance netbox_deltimer(&(flows[ndx].timer));
        NETBOX_Instance netbox_newtimer(&(flows[ndx].timer), _VOID_VOID_ flow_timeout, flows+ndx, RESV_TIMEOUT);
        return 0;
}*/

void NS_CLASS flow_timeout(void* vp)
{
        struct flow_s *n=(struct flow_s *)vp;
                                    
        if(n->status == 0) {
               //fprintf(stderr, "the status is already 0, so return\n");
               return;
        }
        if(n->status != 3 ) {
               n->status = 1;
               return;
        }
        /*if(n->status <= 2) {
               n->status = 1;
               return;
        }*/
        //n->status = 1;
        //return;
        /*if( n->status != 3) {
               n->status = 0;
               n->bw = 0;
               n->last_bw = 0;
               fprintf(stderr, "node flow_timeout:change the status to 0\n");
        }
	n->last_bw = 0;
        return;*/
}

int NS_CLASS install_peer(struct in_addr ia, sapfsel_t ptr,
			unsigned char plen, char timeroffs)
{
	int ndx; 
	unsigned long t;
        struct in_addr myip;
        
	DEBUG(10, printk("install peer %08x\n", ntohl(ia.s_addr)));
        NETBOX_Instance netbox_getL3addr(&myip, lunardev);
	ndx = add_peer(ia);
	if (ndx < 0) {
		DEBUG(10, printk("install peer: could not add peer\n"));
		return -1;
	}
	if (peers[ndx].used == -1) { // must be in "resolving state"
		DEBUG(10, printk("install peer: not in resolving state\n"));
		return -1;
	}

	peers[ndx].used = -1;     // now resolved
        //peers[ndx].seq = 0;       //sequence number reset to 0
	peers[ndx].hops = plen + 1;
	peers[ndx].flags &= ~PEER_BIDIR;
	peers[ndx].fwdsel = ptr;
        //printf("node %d found a route to %d -- (Hops: %d)\n",myip.s_addr, ia.s_addr, plen+1);
        //fprintf(stderr, "install peer: selector for data is 0x%016Lx\n",ptr);

	t = (timeroffs ? conf_fri/2 : conf_fri) -
		((int) SAPF_Instance getRandomByte()-128) *
		conf_fri_jitter/128;

	NETBOX_Instance netbox_deltimer(&(peers[ndx].timer));
	NETBOX_Instance netbox_newtimer(&(peers[ndx].timer), _VOID_VOID_ peer_timeout,
					peers + ndx, t);

	// if (peers[ndx].flags & PEER_ARP_PENDING) {
	//	peers[ndx].flags &= ~PEER_ARP_PENDING;
	if (peers[ndx].ia.s_addr != htonl(LUNAR_IP_BCAST))
		arp_answer(ndx);
	// }

	return 0;
}
// ----------------------------------------------------------------------
//
 void NS_CLASS unicast_rrep_handler(sapfpkt *p, void *ups_)
{
  struct uc_pending_s *ups = (struct uc_pending_s *)ups_;
	struct xrp_head_s *x = (struct xrp_head_s*) p->data;
	struct xrp_sel_eth_s *f;
        struct xrp_bw_s *b = 0;

        DEBUG(9, printk("unicast_rrep_handler (%d bytes)\n", p->len));
        //fprintf(stderr, "Here is unicast_rrep_handler\n");
	if (p->len < sizeof(struct xrp_head_s) ||
	    x->version != XRP_VERSION_REPLY) {
		DEBUG(9, printk("  unknown or garbled RREP\n"));
		return ;
	}
         
        // The following lines were added by me**************
        b = (struct xrp_bw_s *) xrp_locate(x+1, p->len -sizeof(*x), XRP_CLASS_BW);
        if (!b || b->partype != XRP_PARTYPE_BW) {
                DEBUG(9, printk(" no bandwidth found in RREP\n"));
                return;
        }
        //fprintf(stderr, " The bw in RREP is %6f\n", b->bw);
        int idx = add_flow(x->src, x->dst);
        if (idx < 0) {
                DEBUG(10, printk(" install flows: could not install flow\n"));
                return;
        }
        //printf("initiating node %d: flow status of %d to %d is %d\n",x->src, x->src,x->dst,flows[idx].status);
        //fprintf(stderr, "source node recv RREP\n");
        // only accept one path, ignore other late RREP
        if ( flows[idx].status == 5 || flows[idx].status == 3 || flows[idx].status == 2) {
                //printf("recv RREP, but the flow status is %d, so ignore\n",flows[idx].status);
                return;
        }
        
        int n = x->ttl + 1;
        if(n <= 2)
                flows[idx].bw += n*(b->bw);
        else flows[idx].bw += 3*(b->bw);
        //flows[idx].status = 2;
        double abw;
        double csn_bw_cur;
        if (flows[idx].status == 4) {
               abw = check_bandwidth()+flows[idx].last_bw;
               csn_bw_cur = csn_bw + flows[idx].last_bw;
               flows[idx].status = 5;
        }
        else {
               // the case status == 2 was moved to the above if statement
               /*if (flows[idx].status == 2) {
                   abw = check_bandwidth()+flows[idx].bw;
                   csn_bw_cur = csn_bw;
               }*/
               if (flows[idx].status == 1) {
                   abw = check_bandwidth();
                   csn_bw_cur = csn_bw;
                   flows[idx].status = 2;
               }

               //else fprintf(stderr, "Source %d recv RREP, but the flow status is %d\n",x->src,flows[idx].status);
        }
        /*if (abw < flows[idx].bw || csn_bw_cur < flows[idx].bw) {
                //printf("source--check bw when RREP aarives, can't meet bw requirement, so exit!!!!!!!!!!!!!!!!!!!\n");
                //flows[idx].status = 0;
                return;
        }*/
	if(abw >= flows[idx].bw && csn_bw_cur >= flows[idx].bw) {
                //NETBOX_Instance netbox_deltimer(&(flows[idx].timer));
                //NETBOX_Instance netbox_newtimer(&(flows[idx].timer), _VOID_VOID_ flow_timeout, flows+idx, RESV_TIMEOUT);
        
	        f = (struct xrp_sel_eth_s *) xrp_locate(b+1, p->len - sizeof(*x) -sizeof(*b), XRP_CLASS_FWD_ADDR);
	        if (!f || f->partype != XRP_PARTYPE_SEL_ETH) {
		      DEBUG(9, printk("  no forward addr found in RREP %p\n", f));
		      return ;
	        }
                //fprintf(stderr, "recved RREP, install peer\n");
	        if (install_peer(ups->ia, mkNetPointer(f->eth, f->sel), x->ttl, 0))
		      return ;
                //fprintf(stderr, "install peer successfully\n");
	        if (ups->is_bidir)
	            SAPF_Instance sapf_addREDIR(mysapfhandle, ups->backsel, ipv4deliver);
	        NETBOX_Instance netbox_deltimer(&(flows[idx].timer));
		NETBOX_Instance netbox_newtimer(&(flows[idx].timer), _VOID_VOID_ flow_timeout, flows+idx, RESV_TIMEOUT);
				
	} else {
                flows[idx].status = 0;
	}

        SAPF_Instance sapf_deleteEntry(mysapfhandle, ups->caller);

        return ;
     
}

void NS_CLASS unicast_transit_rrep_handler(sapfpkt *p,void *tr_)
{
        struct uc_transit_pending_s *tr =(struct uc_transit_pending_s *)tr_;
	int len, ifnr = basedev ? basedev->ifindex : 0;
	struct xrp_head_s *x;
	struct xrp_sel_eth_s *f;
        struct xrp_bw_s *b;
	struct sockaddr sa;
	char buf[300];
        struct in_addr myip;
        NETBOX_Instance netbox_getL3addr(&myip, lunardev);
        DEBUG(9, printk("unicast transit rrep handler (%d bytes)\n",
			p->len));
        //fprintf(stderr, "node %d recv RREP (%d -> %d)\n",myip.s_addr, x->src,x->dst);
	if (p->len < sizeof(struct xrp_head_s))
		return ;

	len = MIN(sizeof(buf), p->len);
	memcpy(buf, p->data, len);
	x = (struct xrp_head_s *) buf;
	if (x->version != XRP_VERSION_REPLY) {
		DEBUG(9, printk("  unknown or garbled RREP\n"));
		return ;
	}
	x->ttl++;
            
        // the following lines were added by me
        b = (struct xrp_bw_s *) xrp_locate(x+1, len-sizeof(*x), XRP_CLASS_BW);
        if (!b || b->partype != XRP_PARTYPE_BW) {
                DEBUG(9, printk(" no bandwidth found in RREP\n"));
                return;
        }        
        int idx = add_flow(x->src, x->dst);
        if (idx < 0) {
                DEBUG(10, printk("install flows: could not install flow\n"));
                return;
        }
        if (flows[idx].status == 3 || flows[idx].status == 5)    //we want to find node-disjoint paths
                return;
        if (x->ttl < 3)
                flows[idx].bw += (x->ttl)*(b->bw);
        else    flows[idx].bw += 3*(b->bw);
        //flows[idx].status = 2;
        double abw;
        double csn_bw_cur;
        if (flows[idx].status == 4) {
                //if (flows[idx].last_bw == 0)
                //    abw = check_bandwidth() + flows[idx]
                abw = check_bandwidth() + flows[idx].last_bw;
                csn_bw_cur = csn_bw + flows[idx].last_bw;
                //flows[idx].status = 5;
        }
        else {
                if (flows[idx].status == 1) {
                    abw = check_bandwidth();
                    csn_bw_cur = csn_bw;
                    //flows[idx].status = 2;
                    //fprintf(stderr, "node %d resv bw %.4f for %d -> %d\n",myip.s_addr,x->src,x->dst);
                }
                if (flows[idx].status == 2) {
                    abw = check_bandwidth()+flows[idx].bw;
                    csn_bw_cur = csn_bw;
                }
                //else fprintf(stderr, "in RREP phase, the flow status is other than 1 or 4\n");
        }
        if (abw > Bmax)
                abw = Bmax;
        if (csn_bw_cur > Bmax)
                csn_bw_cur = Bmax;
        /*if (abw < flows[idx].bw || csn_bw_cur < flows[idx].bw) {
                //sprintf("intermediate node(RREP) %d: can't meet bw requirement between node %d and %d --exit!!!!!!!\n", myip.s_addr,x->src, x->dst);
                //flows[idx].status = 0;
                return;
        }*/
        //printf("node %d forward Route Reply from %d\n",myip.s_addr, x->src);
	if(abw > flows[idx].bw && csn_bw_cur > flows[idx].bw) {
                if(flows[idx].status == 4)
                        flows[idx].status = 5;
                if (flows[idx].status == 1) {
                        flows[idx].status = 2;
                }
                //flows[idx].status = 2;
                //fprintf(stderr, " intermediate node recv RREP\n");
                //fprintf(stderr, " In RREP, node %d: the avail bw is %6f, and the request bw is %6f\n",myip.s_addr,abw, flows[idx].bw);
                NETBOX_Instance netbox_deltimer(&(flows[idx].timer));
                NETBOX_Instance netbox_newtimer(&(flows[idx].timer), _VOID_VOID_ flow_timeout, flows+idx, RESV_TIMEOUT);
                //**************************************************************
        
	        f = (struct xrp_sel_eth_s *)
		//xrp_locate(x+1, len - sizeof(*x), XRP_CLASS_FWD_ADDR);
                xrp_locate(b+1, len - sizeof(*x) - sizeof(*b), XRP_CLASS_FWD_ADDR);
	        if (!f || f->partype != XRP_PARTYPE_SEL_ETH) {
		       DEBUG(9, printk("  no forward addr found in RREP %p\n", f));
		       return ;
	        }

	        if (tr->is_bidir)
		SAPF_Instance sapf_addCallback(mysapfhandle, tr->mybacksel, NS_CLASS_addr L2sendto,
		mkSapfDest(tr->backeth, tr->backsel));

	        NETBOX_Instance netbox_getL2addr(ifnr, &sa);
	        f->sel = mkNetPointer(f->eth, f->sel);
	        memcpy(f->eth, sa.sa_data, ETH_ALEN);
                //fprintf(stderr, "node %d forward reply for %d -> %d, next hop is %d\n",myip.s_addr,x->src,x->dst,f->eth);

	        memcpy(sa.sa_data, tr->replyeth, ETH_ALEN);
	        NETBOX_Instance netbox_l2send(ifnr, &sa, tr->replysel, buf, len);

                SAPF_Instance sapf_deleteEntry(mysapfhandle, tr->caller);
	}
        return ;
}

 void NS_CLASS bcast_rep_handler(sapfpkt *p, void *bps_)
{
  struct bcast_pending_s *bps=(struct bcast_pending_s *)bps_;
	struct xrp_head_s *x = (struct xrp_head_s*) p->data;
	struct xrp_sel_eth_s *f;
	//struct xrp_common_s *f; //changed by me 
	NS_STATIC unsigned char broadcast[6] = {'\xff', '\xff', '\xff', '\xff', '\xff', '\xff'};

        DEBUG(9, printk("bcast_rep_handler (%d bytes)\n", p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
	    x->version != XRP_VERSION_REPLY) {
		DEBUG(9, printk("  unknown or garbled REP\n"));
		return ;
	}
	f = (struct xrp_sel_eth_s *)xrp_locate(x+1, p->len - sizeof(*x), XRP_CLASS_FWD_ADDR);
	//f = (struct xrp_common_s *)xrp_locate(x+1, p->len - sizeof(*x), XRP_CLASS_FWD_ADDR);   //changed by me
	if (!f || f->partype != XRP_PARTYPE_SEL_ETH) {
		DEBUG(9, printk("  no forward addr found in RREP %p\n", f));
		return ;
	}
	DEBUG(10, printk(" f->sel = 0x%016Lx\n; bps->fwd_sel = 0x%016Lx\n", f->sel, bps->fwd_sel));
	//uint8_t *f8 = (uint8_t *)f;  //added by me
	if (f->sel != bps->fwd_sel) {
	//if (memcmp(f8+4, &(bps->fwd_sel), sizeof(sapfsel_t))) {
		DEBUG(10, printk(" f->sel = 0x%016Lx; bps->fwd_sel = 0x%016Lx\n", f->sel, bps->fwd_sel));
		DEBUG(9, printk("  returned bc sel does not match\n"));
		return ;
	}

	bc_sel = mkNetPointer(broadcast, f->sel); 
	//sapfsel_t bsel;
	//memcpy(&bsel, f8+4, sizeof(bsel));
	//bc_sel = mkNetPointer(broadcast, bsel);
	DEBUG(10, printk(" bcast_rep_handler -- bps->ia.addr = 0x%8x\n", bps->ia.s_addr));
	if (!bc_len) {
		install_peer(bps->ia, bc_sel, 3, 0);
		if (bc_selp)
			*bc_selp = bc_sel;
	}

        SAPF_Instance sapf_deleteEntry(mysapfhandle, bps->caller);

        return ;
}

// ----------------------------------------------------------------------

void NS_CLASS lunar_rreq_handler(sapfpkt *p, void *aux)
{
	int ifnr = basedev ? basedev->ifindex : 0;
	struct xrp_head_s *x = (struct xrp_head_s *) p->data;
	struct xrp_sel_eth_s *reply = 0, *backwd = 0;
	struct xrp_sel_s *series = 0, *fwd = 0;
	struct xrp_ipv4_s *src = 0, *target = 0;
        struct xrp_bw_s *bandwidth = 0;
	struct in_addr myip;
	struct sockaddr sa;
	xrp_param_t *pp; // pointer to a single parameter block
	int cnt, len;
        //fprintf(stderr, "Here is lunar_rreq_handler\n");
	DEBUG(9, printk("LUNAR: RREQ (%d bytes)\n", p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
	    x->version != XRP_VERSION_REQUEST) {
                //fprintf(stderr, " Wrong length or version\n");
		DEBUG(9, printk(" wrong length or version\n"));
		return;
	}

	len = p->len - sizeof(*x);
	pp = (xrp_param_t *) (x+1);
        
	while (len >= 4) {
		if (pp->param.len == 0)
			break;
		if (ntohs(pp->param.len) > len ||
		    ntohs(pp->param.len) < 4) {
                        
			DEBUG(9, printk(" end of XRP parsing: wrong length\n"));
			return;
		}
		if (pp->param.partype > XRP_PARTYPE_UNKNOWN &&
		    pp->param.partype < XRP_PARTYPE_COUNT) {
			if (ntohs(pp->param.len) !=
			    xrp_param_len[pp->param.partype]) {
				DEBUG(9, printk(" pp->param.len = %d; xrp_param_len = %d \n", ntohs(pp->param.len), xrp_param_len[pp->param.partype]));
				DEBUG(9, printk(" wrong parameter length, class=%d\n", pp->param.parclass));
                                return;
			}
			DEBUG(10, printk(" param class+len+expect_len: %d %d %d\n",
			pp->param.parclass, ntohs(pp->param.len), xrp_param_len[pp->param.partype]));
                        
			switch (pp->param.parclass) {
                        //the following lines are added by me
                        case XRP_CLASS_BW:
                                if (pp->param.partype == XRP_PARTYPE_BW)
                                        bandwidth = (struct xrp_bw_s *) pp;
                                break;        
                        //**************************************
			case XRP_CLASS_REQ_SERIES:
                                if (pp->param.partype == XRP_PARTYPE_SEL)
					series = (struct xrp_sel_s *) pp;
				break;
			case XRP_CLASS_REPLY_ADDR:
                                if (pp->param.partype == XRP_PARTYPE_SEL_ETH)
					reply = (struct xrp_sel_eth_s *) pp;
				break;
			case XRP_CLASS_TARGET:
                                if (pp->param.partype == XRP_PARTYPE_IPv4)
					target = (struct xrp_ipv4_s*) pp;
				break;
			case XRP_CLASS_BACKWD_ADDR:
                                //fprintf(stderr, " parclass is XRP_CLASS_BACKWD_ADDR\n");
				if (pp->param.partype == XRP_PARTYPE_SEL_ETH)
					backwd = (struct xrp_sel_eth_s *) pp;
				break;
			case XRP_CLASS_FWD_ADDR: // for broadcast
                                 //fprintf(stderr, " parclass is XRP_CLASS_FWD_ADDR\n");
				 if (pp->param.partype == XRP_PARTYPE_SEL)  
				 //if (pp->param.partype == XRP_PARTYPE_SEL_ETH)
					fwd = (struct xrp_sel_s *) pp;
				break;
			case XRP_CLASS_SRC_ADDR:
                                //fprintf(stderr, " parclass is XRP_CLASS_SRC_ADDR\n");
				if (pp->param.partype == XRP_PARTYPE_IPv4)
					src = (struct xrp_ipv4_s *) pp;
				break;
			default: // ignore unknown fields
                                 //fprintf(stderr, " parclass is UNKNOW\n");
				break;
			}
		}
		cnt = (ntohs(pp->param.len) + 3) & ~0x3;
		len -= cnt;
		pp = (xrp_param_t*) (((char*)pp) + cnt);
	}
        
	if (!series || SAPF_Instance sapf_exists(mysapfhandle, series->sel)) {
		DEBUG(9, printk(" no series/series %016Lx already exists, dropping\n", series ? ntohsel(series->sel) : const64(0)));
                //fprintf(stderr, "The selector already exist\n");
		return;
	}

        
        //*************************************************       

	//SAPF_Instance sapf_addREDIR(mysapfhandle, series->sel, 0);

	NETBOX_Instance netbox_getL3addr(&myip, lunardev);
	NETBOX_Instance netbox_getL2addr(ifnr, &sa);

	if (!target || !reply) {
		DEBUG(9, printk(" no target or no reply sel: %p %p\n",
				target, reply));
		return;
	}
	DEBUG(9, printk("  target is IP 0x%08x, I'm 0x%08x\n", ntohl(target->ip), ntohl(myip.s_addr)));
	DEBUG(10, printk(" LUNAR_IP_BCAST is 0x%08x\n", htonl(LUNAR_IP_BCAST)));

	if (target->ip == htonl(LUNAR_IP_BCAST)) {
		unsigned char buf[300], *cp = buf;
		sapfsel_t ptr;

		DEBUG(1, printk("broadcast tree request received\n"));
                SAPF_Instance sapf_addREDIR(mysapfhandle, series->sel, 0);
		if (!fwd) {
			DEBUG(9, printk(" no fwd sel\n"));
			return;
		}
		//ptr = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, fwd->sel);
		ptr = fwd->sel;
		DEBUG(10, printk(" lunar_rreq_handle: the ptr for bcast is 0x%016Lx, 0x%016Lx\n", fwd->sel,ptr));
		SAPF_Instance sapf_addREDIR(mysapfhandle, ptr, ipv4deliver);

		cp += xrp_fill_hdr(cp, XRP_VERSION_REPLY, 0, 0, 0, 0);
                cp += xrp_fill_bw(cp, XRP_CLASS_BW, bandwidth->bw);
		//int ll = xrp_fill_sel_eth(cp, XRP_CLASS_FWD_ADDR, ptr, (unsigned char *)sa.sa_data);
		//DEBUG(10, printk(" xrp_fill_sel_eth returns length %d\n", ll));
		//cp += ll;
		cp += xrp_fill_sel_eth(cp, XRP_CLASS_FWD_ADDR, ptr, (unsigned char *)sa.sa_data);
		unsigned char hid[NETBOX_HID_LEN];
          
		sapfsel_t bsel;
		DEBUG(10, printk("IP data :" ));
		int i;
		for ( i = 0; i < 50; i++) {
			DEBUG(10, printk(" %02x", buf[i]));
		}	
		DEBUG(10, printk("\n"));

		
						
		//DEBUG(10, printk(" the sel in the reply packet is 0x%128Lx\n", buf+20));
		memcpy(&bsel, buf+28, sizeof(bsel));
		DEBUG(10, printk(" the sel in the reply packet is 0x%016Lx\n", bsel)); 

		if (!NETBOX_Instance netbox_getHID128(hid))
			cp += xrp_fill_hid128(cp, XRP_CLASS_SRC_ID, hid);
		*cp++ = 0; *cp++ = 0;

		// ***********************test*************************************************
		struct xrp_sel_eth_s *f;
		f = (struct xrp_sel_eth_s *)xrp_locate(buf+sizeof(struct xrp_head_s), cp-buf-sizeof(struct xrp_head_s), XRP_CLASS_FWD_ADDR);
		if (!f || f->partype != XRP_PARTYPE_SEL_ETH) {
		    DEBUG(9, printk("  no forward addr found in RREP %p\n", f));
		    return ;
		}
		DEBUG(10, printk(" f->sel = 0x%016Lx\n", f->sel));
		//************************************************************************************
		memcpy(sa.sa_data, reply->eth, ETH_ALEN);
		NETBOX_Instance netbox_l2send(ifnr, &sa, reply->sel, buf, cp - buf);
                DEBUG(10, printk("Send back reply upon receiving bcast req\n"));
		return;
	}
  
         
	if (target->ip == myip.s_addr && x->ttl >= 1) { // unicast match
                //fprintf(stderr, "the target ip is my ip %d-- unicast match\n",myip.s_addr);
                int idx = add_flow(x->src, x->dst);
                if (idx < 0) {
                     DEBUG(10, printk("refresh: no such entry in flow table\n"));
                     return;
                }
                double abw;
                double csn_bw_cur;
                //printf("%d is the target, my status for the flow is %d\n",myip.s_addr,flows[idx].status);
                //if (flows[idx].status == 4) {
                if (flows[idx].status == 3 || flows[idx].status == 5) {
                     //fprintf(stderr, "dst node check_bandwidth (%d -> %d) = %.4f\n",flows[idx].src,flows[idx].dst,check_bandwidth());
                     abw = check_bandwidth()+flows[idx].last_bw;
		     //abw = check_bandwidth();
                     csn_bw_cur = csn_bw + flows[idx].last_bw;
                     //flows[idx].status = 5;     
                } else {
                     if (flows[idx].status == 2) {
                          //abw = check_bandwidth() + flows[idx].bw;
			  abw = check_bandwidth();
                     } else {
                             abw = check_bandwidth();
                     }
                     csn_bw_cur = csn_bw;
                    
                }
                if (abw > Bmax)
                           abw = Bmax;
                if (csn_bw_cur > Bmax)
                           csn_bw_cur = Bmax;
                                                                                
                int n = x->hops + 1;
                double rbw;
                if (n < 2 )
                     rbw = n*(bandwidth->bw);
                else    rbw = 2*(bandwidth->bw);
                /* if (abw < rbw || csn_bw_cur < rbw) {
                 //fprintf(stderr, "now in RREQ, the bandwidth not meet the requirement -- exit !!!!\n");
                   //printf("dst node %d: local bw is %6f; csn bw is %6f; requested bw is %6f -- reject the request of node %d for %d !!!! \n",myip.s_addr,abw, csn_bw_cur, rbw, x->src, x->dst);
                   //comment this line because the dst can accept multiple RREQs
                        //flows[idx].status = 0;
                        return;
                 }*/
	        if(abw > rbw && csn_bw_cur > rbw) {
                        flows[idx].bw = rbw;
                        if (flows[idx].status == 3 || flows[idx].status == 5)
                               flows[idx].status = 5;
                        else {
                               flows[idx].status = 2;    // bandwidth reserved
                        }
                        NETBOX_Instance netbox_deltimer(&(flows[idx].timer));
                        NETBOX_Instance netbox_newtimer(&(flows[idx].timer), _VOID_VOID_ flow_timeout, flows+idx, RESV_TIMEOUT);


                        //printf(" I am the target, recv RREQ from %d, send back RREP\n",flows[idx].src );
		        unsigned char buf[300], *cp = buf;
		        unsigned char hid[NETBOX_HID_LEN];
		        sapfsel_t ptr = SAPF_Instance sapf_getUnusedSelector(mysapfhandle,
						       SELCLASS_LOCAL);

		        DEBUG(10, printk("LUNAR: I was discovered\n"));
                         
                        //*****************************************************

		        // fprintf(stderr, "backwd=%p src=%p, %08x\n", backwd, src, src ? src->ip : 0);
		        if (backwd && src && peer_is_unknown(src->ip)) {
			         struct in_addr ia;
			         ia.s_addr = src->ip;
			         install_peer(ia, mkNetPointer(backwd->eth, backwd->sel), 9, 1);
		        }

		        SAPF_Instance sapf_addREDIR(mysapfhandle, ptr, ipv4deliver);
                                                   
		        cp += xrp_fill_hdr(cp, XRP_VERSION_REPLY, 0, 0, x->src, myip.s_addr);
                        cp += xrp_fill_bw(cp, XRP_CLASS_BW, bandwidth->bw);

                        cp += xrp_fill_sel_eth(cp, XRP_CLASS_FWD_ADDR, ptr, (unsigned char*)sa.sa_data);
                	if (!NETBOX_Instance netbox_getHID128(hid))
		        	cp += xrp_fill_hid128(cp, XRP_CLASS_SRC_ID, hid);
		        *cp++ = 0; *cp++ = 0;
                        memcpy(sa.sa_data, reply->eth, ETH_ALEN);
                	NETBOX_Instance netbox_l2send(ifnr, &sa, reply->sel, buf, cp - buf);
		}
	} else if (x->ttl-- > 1) { // re-broadcast the request
 		struct uc_transit_pending_s *tr;
                //fprintf(stderr, "now in RREQ for %d -- node %d is not target, forwading\n", target->ip, myip.s_addr);
                SAPF_Instance sapf_addREDIR(mysapfhandle, series->sel, 0);
                int idx = add_flow(x->src, x->dst);
                if (idx < 0) {
                      //printf("refresh: no such entry in flows\n");
                      return;
                }
                double abw;
                double csn_bw_cur;
                //if (flows[idx].status == 3 || flows[idx].status == 4) {
                if (flows[idx].status >=3) {
                      flows[idx].status = 4; //refreshing route discovery
                      //fprintf(stderr, "node %d check_bandwidth: %.4f\n",myip.s_addr,check_bandwidth());
                      abw = check_bandwidth()+flows[idx].last_bw;
                      csn_bw_cur = csn_bw + flows[idx].last_bw;
                      if (abw > Bmax)
                           abw = Bmax;
                      if (csn_bw_cur > Bmax)
                           csn_bw_cur = Bmax;
                } else {
                       //this is the temperory strategy, in case that when an intermediate node already returned a RREP and change status to 2, but the source send out a RREQ again.if we don'tuse this strategy, the status at intermediate node will change back to 1 and can't forward data   
                       if (flows[idx].status == 2) {
                           abw = check_bandwidth() + flows[idx].bw;
                           flows[idx].status = 2;
                       }
                       else {
                           flows[idx].status = 1;
                       //fprintf(stderr, "%d in RREQ (%d to %d), change status to 1\n",myip.s_addr, flows[idx].src, flows[idx].dst);
                           abw = check_bandwidth()+flows[idx].last_bw;
                           //abw = check_bandwidth();
                       }
                       if (abw > Bmax)
                           abw = Bmax;
                      //}
                      //else {
                      //     abw = check_bandwidth();
                      //     flows[idx].status = 1;
                      //}
                      csn_bw_cur = csn_bw;
                 }                                                          
                int n = x->hops + 1;
                double rbw;
                if (n < 2 )
                      rbw = n*(bandwidth->bw);
                else    rbw = 2*(bandwidth->bw);
                if (abw < rbw || csn_bw_cur < rbw) {
                      //fprintf(stderr, "node %d (RREQ): local bw is %6f; csn bw is %6f; requested bw is %6f -- reject the request of node %d for %d !!!! \n",myip.s_addr,abw, csn_bw_cur,rbw, x->src, x->dst);
                      //flows[idx].status = 0;
                      return;
                 }
                 //fprintf(stderr,"node %d forward RREQ for %d -> %d\n",myip.s_addr,x->src,x->dst);
                //SAPF_Instance sapf_addREDIR(mysapfhandle, series->sel, 0);
                
                flows[idx].bw = rbw;
                //flows[idx].status = 1;
                long tout = (conf_max_retries-conf_max_ring+1)*conf_max_ring*conf_timeout_per_ring; 
                NETBOX_Instance netbox_deltimer(&(flows[idx].timer));
                NETBOX_Instance netbox_newtimer(&(flows[idx].timer), _VOID_VOID_ flow_timeout, flows+idx, tout); 

                //*********************************************

		if (!(tr = (struct uc_transit_pending_s *)kmalloc(sizeof(*tr), GFP_ATOMIC)))
			return;
		memset(tr, 0, sizeof(*tr));

		tr->replysel = reply->sel;
		memcpy(tr->replyeth, reply->eth, ETH_ALEN);

		memcpy(reply->eth, sa.sa_data, ETH_ALEN);
		tr->caller = reply->sel = mkHandlerPointer(
			NS_CLASS_addr unicast_transit_rrep_handler, tr);
		if (backwd) {
			tr->is_bidir = 1;
			tr->backsel = backwd->sel;
			memcpy(tr->backeth, backwd->eth, ETH_ALEN);

			tr->mybacksel = backwd->sel =
				SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
			memcpy(backwd->eth, sa.sa_data, ETH_ALEN);
		}
                x->hops = x->hops + 1;
		schedule_rreq(p->data, p->len, 0);
	} // else drop

	return;
}


void NS_CLASS send_rreq(int radius, sapfsel_t series, uint32_t target, 
                      sapfsel_t reply, sapfsel_t backsel, int32_t src,
                      sapfsel_t fwd, uint8_t hops, int32_t source, double bw)
{
	struct sockaddr sa;
	int ifnr = basedev ? basedev->ifindex : 0;
	char buf[300], *cp = buf;
        cp = buf;
	cp += xrp_fill_hdr(cp, XRP_VERSION_REQUEST, radius, hops, source,target);
        cp += xrp_fill_bw(cp, XRP_CLASS_BW, bw);
	cp += xrp_fill_sel(cp, XRP_CLASS_REQ_SERIES, series);
	NETBOX_Instance netbox_getL2addr(ifnr, &sa);
	cp += xrp_fill_sel_eth(cp, XRP_CLASS_REPLY_ADDR,
			       reply, (unsigned char *)sa.sa_data);
	cp += xrp_fill_ipv4(cp, XRP_CLASS_TARGET, target);

	if (src != NO_IP_ADDR) {
		cp += xrp_fill_ipv4(cp, XRP_CLASS_SRC_ADDR, src);
		cp += xrp_fill_sel_eth(cp, XRP_CLASS_BACKWD_ADDR,
			       backsel, (unsigned char *)sa.sa_data);
	}
	// else fprintf(stderr, "send_rreq: no src\n");

	if (fwd) {
		cp += xrp_fill_sel(cp, XRP_CLASS_FWD_ADDR, fwd);
	}

	*cp++ = 0; *cp++ = 0;
        
	NETBOX_Instance netbox_l2send(ifnr, 0,
				      htonsel(SAPF_SELNET_LUNAR_RREQ),
				      buf, cp - buf);
}

 void NS_CLASS start_rreq(int ndx)
{
	//fprintf(stderr, "Here is start_rreq\n");
        sapfsel_t series;
	int radius, i;
	long tout;
        struct in_addr myip;

	DEBUG(10, printk("going to start_rreq\n"));
        
        NETBOX_Instance netbox_getL3addr(&myip, lunardev);
	DEBUG(10, printk("start_rreq called for %08x\n", peers[ndx].ia.s_addr));
        //printf("node %d start_rreq called for %d\n", myip.s_addr, peers[ndx].ia.s_addr);
     
        //*******************************************************

	radius = MIN(peers[ndx].used, conf_max_ring);
	series = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
	SAPF_Instance sapf_addREDIR(mysapfhandle, series, 0);

	if (peers[ndx].ia.s_addr != htonl(LUNAR_IP_BCAST)) {
		struct uc_pending_s *ups;
                int idx = add_flow(myip.s_addr, peers[ndx].ia.s_addr);     
                if (idx < 0) {
                      //fprintf(stderr, "refresh: no such entry in flows\n");
                      return;
                }
                double abw;
                double csn_bw_cur;
                //printf("node %d start_rreq: flow status is %d\n",myip.s_addr,flows[idx].status);
                 //if (flows[idx].status == 3 || flows[idx].status == 4) {
                if (flows[idx].status >= 3) {
                       flows[idx].status = 4; //refreshing route discovery
                       abw = check_bandwidth()+flows[idx].last_bw;
		       //abw = check_bandwidth();
                       csn_bw_cur = csn_bw + flows[idx].last_bw;
                       if (abw > Bmax)
                              abw = Bmax;
                       if (csn_bw_cur > Bmax)
                              csn_bw_cur = Bmax;
                 }
                 else {
                       flows[idx].status = 1;
                       abw = check_bandwidth();
                       csn_bw_cur = csn_bw;
                 }
                
                if (abw < peers[ndx].bw || csn_bw_cur < peers[ndx].bw) {
                        DEBUG(10, printk("request rejected\n"));
                        //printf("source node %d: local bw is %6f; csn bw is %6f; the req bw is %6f;reject the request !!!!!\n", myip.s_addr,abw, csn_bw_cur,peers[ndx].bw);
                        //flows[idx].status = 0;
			radius = 0;
			DEBUG(10, printk("at the source node, bw can't be met, the radius is set 0\n"));
                        //return;
                }
                //fprintf(stderr, "Source node %d: local bw is %4f; csn bw is %4f; requested bandwidth is %4f\n",myip.s_addr,abw,csn_bw_cur,peers[ndx].bw);
                //fprintf(stderr, "accept the route request\n");
             
                DEBUG(10, printk("request accepted\n"));
                //NETBOX_Instance netbox_getL3addr(&myip, lunardev);
                //int idx = add_flow(myip.s_addr, peers[ndx].ia.s_addr);
                //fprintf(stderr, "start rreq -- the available position in flows is %d\n", idx);
                //if (idx < 0) {
                //        DEBUG(10, printk("install flows: could not install flow\n"));
                //        return;
                //} 
                //fprintf(stderr, "start_rreq: add flow %d, the src and dst are %d and %d\n -- ", idx, myip.s_addr, peers[ndx].ia.s_addr);
                flows[idx].bw = 0;
                //flows[idx].status = 1;
                //fprintf(stderr, "the added flow's %d status is %d\n",idx, flows[idx].status);
                //fprintf(stderr, "Successfully add a flow\n");
                //************************************************

		ups = (struct uc_pending_s *)kmalloc(sizeof(struct uc_pending_s), GFP_ATOMIC);
		if (!ups) {
			peers[ndx].used = 0;
			return;
		}
		memset(ups, 0, sizeof(struct uc_pending_s));
		ups->ia = peers[ndx].ia;
		ups->caller = mkHandlerPointer(
			NS_CLASS_addr unicast_rrep_handler, ups);

		if (peers[ndx].flags & PEER_BIDIR) {
			//struct in_addr myip;

			ups->is_bidir = 1;
			ups->backsel = peers[ndx].backsel;
			//NETBOX_Instance netbox_getL3addr(&myip, lunardev);
			send_rreq(radius, series, peers[ndx].ia.s_addr,
				  ups->caller, ups->backsel, myip.s_addr, 0, 0, myip.s_addr, peers[ndx].bw);
		} else
			send_rreq(radius, series, peers[ndx].ia.s_addr,
				  ups->caller, 0, NO_IP_ADDR, 0, 0,myip.s_addr, peers[ndx].bw);
                
		tout = radius * conf_timeout_per_ring;
                i = peers[ndx].used;
		NETBOX_Instance netbox_deltimer(&(flows[idx].timer));
                //NETBOX_Instance netbox_newtimer(&(flows[idx].timer), _VOID_VOID_ flow_timeout, flows+idx, tout);

	} else { // set up broadcast tree

		struct bcast_pending_s *bps;

		bps = (struct bcast_pending_s *)kmalloc(sizeof(struct bcast_pending_s), GFP_ATOMIC);
		if (!bps) {
			peers[ndx].used = 0;
			return;
		}
		memset(bps, 0, sizeof(struct bcast_pending_s));
		bps->caller = mkHandlerPointer(
			NS_CLASS_addr bcast_rep_handler, bps);
		bps->fwd_sel = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
		bps->ia = peers[ndx].ia;
		DEBUG(10, printk( " start_rreq -- bps->ia.s_addr = 0x%8x\n", bps->ia.s_addr));
		send_rreq(radius, series, bps->ia.s_addr,
			  bps->caller, 0, NO_IP_ADDR, bps->fwd_sel,0,0,0);
		if (bc_len)
			tout = LUNAR_BRCAST_DELAY;
		else
			tout = conf_fri;
	}

	NETBOX_Instance netbox_newtimer(&(peers[ndx].timer), _VOID_VOID_ peer_timeout,
			peers + ndx, tout);
}


 void NS_CLASS peer_timeout(void *vp)
{
  struct in_addr myip;
  NETBOX_Instance netbox_getL3addr(&myip, lunardev);
  struct peer_s *n=(struct peer_s *)vp;
	DEBUG(9, printk("peer_timeout %08x (used=%d)\n",
			ntohl(n->ia.s_addr), n->used));

	if (n->used == 0)
		return;

	if (n->used == -1) {
		if (SAPF_Instance sapf_testandreset_active(mysapfhandle, n->fwdsel)) {
			DEBUG(10, printk("restarting rreq because of activity\n"));
                        //fprintf(stderr, "node %d refreshing route discovery for node %d\n",myip.s_addr,n->ia.s_addr);
			if (n->ia.s_addr == htonl(LUNAR_IP_BCAST))
				n->used = (short) conf_max_retries;
			else { 
                                n->used = (short) conf_min_ring;
                        }
                        start_rreq(n - peers);
		} else {
			DEBUG(10, printk("nonactive peer timeout\n"));
			n->used = 0;
			SAPF_Instance sapf_deleteEntry(mysapfhandle, n->fwdsel);
                        int idx = add_flow(myip.s_addr, n->ia.s_addr);
                        if (idx < 0) {
                           //fprintf(stderr, "refresh: no such entry in flows\n");
                           return;
                        }
                        //printf("no activity recently, change flow (%d -> %d)status to 0\n",myip.s_addr, n->ia.s_addr);
                        //flows[idx].status = 0;
			if (n->ia.s_addr == htonl(LUNAR_IP_BCAST)) {
				DEBUG(10, printk("  bc_selp %p\n", bc_selp));
				bc_sel = 0;
				if (bc_selp) {
					*bc_selp = 0;
					bc_selp = 0;
				}
			} else
				NETBOX_Instance netbox_arpdelete(n->ia, lunardev);
		}
		return;
	}

	if (n->ia.s_addr == htonl(LUNAR_IP_BCAST)) {
		DEBUG(9, printk("bc_delayed_start\n"));

		if (bc_len) {
			if (bc_sel) {
				sapfpkt pkt;

				pkt.data = (char*) bc_buf;
				pkt.len = bc_len;
				SAPF_Instance sapf_demux(mysapfhandle, bc_sel, &pkt);
			}
			bc_len = 0;
		}
		if (bc_selp)
			*bc_selp = bc_sel;
		if (bc_sel)
			install_peer(n->ia, bc_sel, 3, 0);
		
		return;
	}

	if (n->used >= conf_max_retries) { // we give up
		DEBUG(9, printk("LUNAR path setup failed\n"));
		n->used = 0;
                //n->seqn = 0;
		NETBOX_Instance netbox_arpdelete(n->ia, lunardev);
                int ndx = add_flow(myip.s_addr, n->ia.s_addr);
                if (ndx < 0) { 
                    //fprintf(stderr, "give up -- no such entry in flows\n");
                    return;
                }
                //printf("the number of retries from %d byond conf_max_retries\n",myip.s_addr);
                //flows[ndx].status = 1;
                
		return;
	}
        //fprintf(stderr, "route discovery timeout--node %d init again for node %d\n",myip.s_addr,n->ia.s_addr);
	if (n->used < conf_max_ring)
		// 1-hop failed, so try immediately max hops
		n->used = (short) conf_max_ring;
	else    // count up to conf_max_retries
		n->used ++;
        //n->seqn ++;
      
        start_rreq(n - peers);
}


// ----------------------------------------------------------------------

int NS_CLASS lunar_do_arp(struct in_addr name, long bw)
{
	int ndx;
        //fprintf(stderr, "Here is lunar_do_arp");
	DEBUG(10, printk("lunar ARP  %08x\n", ntohl(name.s_addr)));
	DEBUG(10, printk("lunar_do_arp: bw = %ld\n", bw));

	ndx = add_peer(name);
	if (ndx == -1)
		return 0;
	if (peers[ndx].used == -1) {
		arp_answer(ndx);
		return 0;
	}
	DEBUG(10,printk("begin to do ARP\n"));
        //peers[ndx].bw = bw;
//	peers[ndx].flags = PEER_ARP_PENDING;
//	netbox_arpannounce(name, lunardev);

	if (peers[ndx].used > 0) // resolution is pending
		return -1;
        peers[ndx].bw = bw;
	peers[ndx].used = (short) conf_min_ring;
        //peers[ndx].seqn = 0;
	peers[ndx].backsel = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
	//peers[ndx].flags |= PEER_BIDIR;
        peers[ndx].flags = 0;
	start_rreq(ndx);

	return -1;
}


 void NS_CLASS lunar_do_bc_setup(sapfsel_t *sel, sapfpkt *pkt)
{
	struct in_addr bc;
	int ndx;

	DEBUG(9, printk("lunar_do_bc_setup called\n"));

	bc.s_addr = htonl(LUNAR_IP_BCAST);
	ndx = add_peer(bc);
	if (ndx == -1 || peers[ndx].used != 0) {
		return;
	}

	memcpy(bc_buf, pkt->data, pkt->len);
	bc_len = pkt->len;
	bc_sel = 0;
	bc_selp = sel;
	
	peers[ndx].used = (short) conf_max_retries;
	peers[ndx].backsel = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
	peers[ndx].flags |= PEER_BCTREE;
	//**********added by me*************
        peers[ndx].bw = 0;
	//**********************************
	start_rreq(ndx);
}

// ----------------------------------------------------------------------

int NS_CLASS lunar_proc_info(char *buf, char **start, off_t offset, int len)
{
	int written = 0, i, cnt;
	struct peer_s *p;

	for (p = peers, i = LUNAR_MAX_PEERS, cnt = 0; i > 0; p++, i--)
		if (p->used)
			cnt++;

	written += sprintf(buf, "# lunar (%d table entries, %d %s)\n", LUNAR_MAX_PEERS, cnt, cnt == 1 ? "peer" : "peers");

#define adjustlen()	{ \
	if (written < offset) { offset -= written; written = 0; } \
	if (len <= written) \
		goto OUT; \
}
	for (p = peers, i = 0; i < LUNAR_MAX_PEERS; i++, p++) {
		char tmp[20];
		uint32_t ip = ntohl(p->ia.s_addr);
		if (!p->used)
			continue;
		// in kernel space we don't have inet_ntoa():
		sprintf(tmp, "%d.%d.%d.%d",
		    0x0ff & (ip >> 24), 0x0ff & (ip >> 16),
		    0x0ff & (ip >> 8),  0x0ff & ip);
		written += sprintf(buf + written, "%-15s  ", tmp);
		if (p->used > 0) {
			written += sprintf(buf + written,
					   "   searching (attempt %d)\n", p->used);
		} else {
			struct sapfdest_s *d;
			unsigned char *cp;

			written += sprintf(buf + written, p->hops == 0x0f ? 
					   " ? " : "%2d ", p->hops+1);

			d = (struct sapfdest_s*)
				SAPF_Instance sapf_getCallbackAux(mysapfhandle, p->fwdsel);
			cp = (unsigned char *)d->dst.sa_data;
			written += sprintf(buf + written,
					   "%02x:%02x:%02x:%02x:%02x:%02x / %016Lx\n",
					   cp[0], cp[1], cp[2], cp[3],
					   cp[4], cp[5], ntohsel(d->sel));
		}
		adjustlen();
	}

OUT:
	*start = buf + offset;
	written -= offset;
	if (written > len)
		return len;
	return (written < 0) ? 0 : written ;
}

/*double NS_CLASS check_bandwidth()
{
	       return (double)Bmax;
}*/


double NS_CLASS check_bandwidth() 
{
        double rsvw = 0;
        int i,j=0;
        for (i = 0; i < LUNAR_MAX_FLOWS; i++)
        {
             if (flows[i].status == 2) {
                  //fprintf(stderr, "node %d resv bw %.4f for %d -> %d\n",My_IP,flows[i].bw,flows[i].src,flows[i].dst);
                  rsvw = rsvw + flows[i].bw;
             }
             //if (flows[i].dst == My_IP && flows[i].status == 4)
             //     rsvw = rsvw + flows[i].bw;
        }
	//Our_Proc_File->read_proc(buffer, NULL, sizeof(int)+1, 0, NULL, NULL); //proc, buffer,sizeof(int)
        Our_Proc_File->read_proc(buffer,NULL,0,0,NULL,NULL);	
        //local_bwcs_cs = strtol(buffer,NULL,0);
        //local_bw_cs = atof(buffer,NULL,0);
        //local_bw_cs = atof(buffer);
        //local_bw_cs = sinmple_strtol(buffer,NULL,0);
        char* pStr=buffer;
        while(*pStr!='X'){
          pStr++;
          if(j>=100)
          {
             printk(KERN_ERR "illegal argument, in /proc/pkt_rate, j=%d\n",j);
             break;
          }
        }
        pStr++;
        local_bw_cs = simple_strtol(pStr,NULL,0);
        DEBUG(10, printk("the local available bandwidth = %ld\n", local_bw_cs)); 
		
        //fprintf(stderr, "node %d: resv bw %.4f for %d -> %d\n",My_IP,rsvw);
        return local_bw_cs-rsvw;
        
}

/*void NS_CLASS check_bandwidth_cs(void *vp)
{
        
        long tout = CHECK_TIME;
        
        //phy_->check_available_bw();
        mac_->start_bw_check();
        NETBOX_Instance netbox_newtimer(&cs_timer, _VOID_VOID_ cs_timeout, (void*) 0, tout);
}*/

//void NS_CLASS cs_timeout(void* vp)
//{
//        RNG r;
//        mac_->end_check_bw();
//        //fprintf(stderr, "Lunar: bandwidth check ends!\n");
//        lbt=mac_->get_lbtime();
//        if(lbt > CHECK_TIME)
//            lbt = CHECK_TIME;
//        csn_btime=mac_->get_csbtime();
//        if(csn_btime > CHECK_TIME)
//            csn_btime = CHECK_TIME;
        
//        long tout = CHECK_PERIOD;
        
//        local_bw_cs = local_bw_cs/4+3*(1-lbt/CHECK_TIME)*Bmax/4;
        
         
//        double csn_bw_tmp = (1-csn_btime/CHECK_TIME)*Bmax;
//        if (csn_bw_tmp < 0)
//                csn_bw_tmp = 0;
//        if (csn_bw_tmp > Bmax)
//                csn_bw_tmp = Bmax;
        
        ////double csn_bw_cur =csn_bw_tmp - (lbt*Bmax/CHECK_TIME) + (lbt/CHECK_TIME)*(csn_btime/CHECK_TIME)*Bmax;
        
        //fprintf(stderr, "Currently measured csn bw is %6f\n",csn_bw_cur);
//        /*if(csn_bw_cur < 0)
//            csn_bw_cur = 0;
//        if(csn_bw_cur > Bmax)
//            csn_bw_cur = Bmax;*/
//        csn_bw = csn_bw/4 + 3*csn_bw_tmp/4;
        ////csn_bw = csn_bw/3 + 2*csn_bw_cur/3;
        //fprintf(stderr, "cs_timeout -- csn_bw is %6f\n",csn_bw);
        //bw_by_cs = (1-bt/CHECK_TIME)*Bmax;
        //bw_cs_data = (1-bt/CHECK_TIME)*Bmax;
                
//        NETBOX_Instance netbox_newtimer(&bw_timer, _VOID_VOID_ check_bandwidth_cs, (void*)0, tout+long(r.uniform(-5000,5000)));
//} 
 
/*void NS_CLASS check_bandwidth_cs(void *vp)
{
        
}

void NS_CLASS cs_timeout(void* vp) {
 
}*/
        
// eof
