/* xrp.c
 *
 *   SAPF kernel module
 *
 *   Parsing and formatting support for XRP (extensible resolution
 *   protocol)
 *
 * Copyright (C) 2003 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 */

#include "../include/xrp.h"
#include "../include/xrp_decl.h"


void* NS_CLASS xrp_locate(void *buf, int len, int parclass)
{
	struct xrp_common_s *cp = (struct xrp_common_s *)buf;
	int bytecnt;

	while (len >= 4) {
                if (ntohs(cp->len) > len || ntohs(cp->len) < 4)
                        return 0;
		if (cp->parclass == parclass) {
			if (cp->partype >= XRP_PARTYPE_COUNT ||
			    ntohs(cp->len) != xrp_param_len[cp->partype])
				return 0;
			return cp;
		}
		bytecnt = (ntohs(cp->len) + 3) & ~0x3;
                len -= bytecnt;
		cp = (struct xrp_common_s*) (((char*)cp) + bytecnt);
        }
	return 0;
}


int NS_CLASS xrp_fill_hdr(void *buf, int version, int ttl, uint8_t hops, int src, int dst)
{
	struct xrp_head_s *x = (struct xrp_head_s *) buf;
	x->version = (uint8_t)version;
	x->ttl = (uint8_t)ttl;
	//x->flags = x->reserved = 0;
        x->flags = 0;
        x->hops = hops;
        x->src = src;
        x->dst = dst;
	return sizeof(*x);
}


int NS_CLASS xrp_fill_sel(void *buf, int parclass, sapfsel_t sel)
{
	struct xrp_sel_s *p = (struct xrp_sel_s *) buf;
	p->len = htons(sizeof(*p));
	p->parclass = (uint8_t)parclass;
	p->partype = XRP_PARTYPE_SEL;
	p->sel = sel;
	return (ntohs(p->len) + 3) & ~3;
}

int NS_CLASS xrp_fill_bw(void *buf, int parclass, double bw)
{
        struct xrp_bw_s *p = (struct xrp_bw_s *) buf;
        p->len = htons(sizeof(short) + 1 + 1 + sizeof(double));
        p->parclass = (uint8_t)parclass;
        p->partype = XRP_PARTYPE_BW;
        p->bw = bw;
        return (ntohs(p->len) + 3) & ~3;
}

int NS_CLASS xrp_fill_sel_eth(void *buf, int parclass, sapfsel_t sel, unsigned char *eth)
{
	struct xrp_sel_eth_s *p = (struct xrp_sel_eth_s *) buf;
	p->len = htons(sizeof(short) + 1 + 1 + sizeof(sapfsel_t) + ETH_ALEN);
	p->parclass = (uint8_t)parclass;
	p->partype = XRP_PARTYPE_SEL_ETH;
	p->sel = sel;
	//memcpy(p->eth, eth, ETH_ALEN);
	memcpy(p->eth, eth, ETH_ALEN); //this line is mine
	return (ntohs(p->len) + 3) & ~3;
}


int NS_CLASS xrp_fill_ipv4(void *buf, int parclass, int ip)
{
	struct xrp_ipv4_s *p = (struct xrp_ipv4_s *) buf;
	p->len = htons(sizeof(short) + 1 + 1 + 4);
	p->parclass = (uint8_t)parclass;
	p->partype = XRP_PARTYPE_IPv4;
	p->ip = ip;
	return (ntohs(p->len) + 3) & ~3;
}


int NS_CLASS xrp_fill_hid128(void *buf, int parclass, void *hid)
{
	struct xrp_hid128_s *p = (struct xrp_hid128_s *) buf;
	p->len = htons(sizeof(short) + 1 + 1 + sizeof(p->hid));
	p->parclass = (uint8_t)parclass;
	p->partype = XRP_PARTYPE_HID128;
	memcpy(p->hid, hid, sizeof(p->hid));
	return (ntohs(p->len) + 3) & ~3;
}

// eof

