/* platform.c */

#include "precomp.h"
#include "platform.h"

/* In the Windows version we need a pointer to the nw
   adapter in order to be able to e.g. create network
   timers */
extern PADAPT GetAdapter();
extern void set_notifier(struct notifier_block *nb);
extern char basedevname[5];

struct net_device base_dev;
struct net_device lunar_dev;
struct net_device_stats single_stats;

void read_lock(struct rwlock_t *rw)
{
  /* TODO: Perhaps implement this spin lock
     locking function later */
}

void read_unlock(struct rwlock_t *lock)
{
  /* TODO: Perhaps implement this spin lock
     unlocking function later */
}

char *inet_ntoa(struct in_addr in)
{
  static char buf[16];
  UCHAR *p = (UCHAR*)&in;

  sprintf(buf,"%u.%u.%u.%u", p[0], p[1], p[2], p[3] );
  return buf;
}

int inet_aton(char *cp, struct in_addr *ap) {
  int dots = 0;
  u_long acc = 0, addr = 0;
  BYTE temp;

  do {
    register char cc = *cp;

    switch (cc) {
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      acc = acc * 10 + (cc - '0');
      break;
    case '.':
      if (++dots > 3) {
	return 0;
      }
      /* Fall through */
    case '\0':
      if (acc > 255) {
	return 0;
      }
      addr = addr << 8 | acc;
      acc = 0;
      break;
    default:
      return 0;
    }
  } while (*cp++);

  /* Normalize the address */
  if (dots < 3) {
    addr <<= 8 * (3 - dots);
  }

  /* Store it if requested */
  if (ap) {
    ap->s_addr = htonl(addr);
  }

  return 1;
}
/* End check these */

void* kmalloc(unsigned int size,int flags) {
  PVOID voidPtr = NULL;

  /* TODO: Perhaps translate the Unix memory
     allocation flags later on */
  if (NdisAllocateMemoryWithTag(&voidPtr, size, 0) ==
      NDIS_STATUS_SUCCESS) {
    return (void*)voidPtr;
  }

  return NULL;
}

void kfree(const void *objp) {
  if(objp != NULL) {
    NdisFreeMemory((PVOID)objp, 0, 0);
  }
}

void kfree_skb(struct sk_buff *skb) {
  /* Relase memory of contained packet data */
  if (skb->nh.raw != NULL) {
    kfree(skb->nh.raw);
    skb->nh.raw = NULL;
  }

  /* Release memory used by the sk_buff itself */
  kfree(skb);
}

VOID timer_function(IN PVOID SystemSpecific1,
		    IN PVOID FunctionContext,
		    IN PVOID SystemSpecific2,
		    IN PVOID SystemSpecific3) {
  /* Just run the timer function provided */
  struct timer_list *timer = (struct timer_list*)FunctionContext;
  timer->function(timer->data);
}

void init_timer(struct timer_list *timer) {

  PADAPT pAdapt = GetAdapter();

  /* TODO: Remember to return the memory allocated
     to each timer */
  PNDIS_MINIPORT_TIMER mp_timer = (PNDIS_MINIPORT_TIMER)
    kmalloc(sizeof(NDIS_MINIPORT_TIMER), 0);

  if (!mp_timer) {
    DbgPrint("--> Platform - Could not allocate memory for timer <--");
    return;
  }

  timer->list.next = timer->list.prev = NULL;

  NdisMInitializeTimer(mp_timer,
		       pAdapt->MiniportHandle,
		       (PNDIS_TIMER_FUNCTION)(timer_function),
		       (PVOID)timer /* function context */);

  /* Store the miniport timer for future use, i.e.
     in add/delete */
  timer->timer = mp_timer;
}

void add_timer(struct timer_list *timer) {
  int timer_diff;

  /* If no timer had been created, return */
  if (!timer || !timer->timer) {
    return;
  }

  /* Note: The expires value in the timer_list is
     in milliseconds, which is what NdisMSetTimer
     expects */
  timer_diff = (timer->expires - jiffies);
  if (timer_diff > 0) {
    NdisMSetTimer(timer->timer, (UINT)timer_diff);
  } else {
    DbgPrint("--> Did not set timer, illegal timeout: %d ms <--", timer_diff);
  }
}

int del_timer(struct timer_list *timer) {

  BOOLEAN timerCancelled;

  /* If no timer had been created, return error */
  if (!timer || !timer->timer) {
    return -1;
  }

  NdisMCancelTimer(timer->timer, &timerCancelled);

  /* Return the memory used by the miniport
     timer wrapped in the timer struct */
  // Note: This is not needed/cannot be done in our approach
  //kfree(timer->timer);
  //timer->timer = NULL;

  if (timerCancelled) {
    return 0;
  } else {
    return -1;
  }
}

struct net_device* init_etherdev(struct net_device *dev, int sizeof_priv) {
  /* We just return a pointer to the LUNAR device */
  /* TODO: Perhaps we need to handle initialization
     of other ethernet device as well */
  return dev_get_by_index(0);
}

void unregister_netdev(struct net_device *dev) {
  /* Empty implementation */
}

/* This is where we store the functions to notify
   when the network card is registered, comes up, is
   unregistered or goes down */
int register_netdevice_notifier(struct notifier_block *nb) {
  /* We store the notifier block for later use */
  set_notifier(nb);
  return 0;
}

/* This is where we remove the functions that were
   notified about the network card status (see above) */
int unregister_netdevice_notifier(struct notifier_block *nb) {
  /* Store NULL as notifier block, which prevents
     further notifications */
  set_notifier(NULL);
  return 0;
}

void netif_start_queue(struct net_device *dev) {
  /* Empty implementation */
}

void netif_stop_queue(struct net_device *dev) {
  /* Empty implementation */
}

/* Search for an interface by index; In the Windows
   version, only returns a common device */
struct net_device* dev_get_by_index(int ifindex) {
  /* Device 0 is the LUNAR device and device 1
     is the base device; For all other indices
     the LUNAR device is returned */
  /* TODO: Perhaps rework this later; has a certain
     ugly hack flavor in this form */
  if (ifindex == 1) {
    base_dev.name[0] = 'B';
    base_dev.name[1] = 'A';
    base_dev.name[2] = 'S';
    base_dev.name[3] = 'E';
    base_dev.name[4] = '\0';
    base_dev.ifindex = 1;
    base_dev.refcnt.counter = 1;
    base_dev.priv = &single_stats;
    base_dev.hard_header_len = ETH_HLEN;
    return &base_dev;
  }

  lunar_dev.name[0] = 'L';
  lunar_dev.name[1] = 'U';
  lunar_dev.name[2] = 'N';
  lunar_dev.name[3] = 'A';
  lunar_dev.name[4] = 'R';
  lunar_dev.name[5] = '\0';
  lunar_dev.ifindex = 0;
  lunar_dev.refcnt.counter = 1;
  lunar_dev.priv = &single_stats;
  lunar_dev.hard_header_len = ETH_HLEN;
  return &lunar_dev;
}

/* Find an interface by name; In the Windows version,
   just returns device 0 if no name is given; If the
   base device is requested, then that one is returned
   (the device with index 1) */
struct net_device* dev_get_by_name(const char *name) {
  /* Note: In order to return a certain base device */
  /* TODO: Perhaps rework this later; has a certain
     ugly hack flavor in this form */
  if (name && strcmp(name, basedevname) == 0) {
    return dev_get_by_index(1);
  }

  /* Default is to ignore the interface name and
     just return the net device with index 0 (i.e.
     the LUNAR device) */
  return dev_get_by_index(0);
}

void dev_put(struct net_device *dev) {
  /* Empty implementation */
}

/* Creates a number of random bytes and returns them
   in a buffer pointed to by buf */
void get_random_bytes(void *buf, int nbytes) {
  int i;
  unsigned char* bytebuf;

  bytebuf = (unsigned char*)buf;
  for(i = 0; i < nbytes; i++) {
    /* Note: We scale down to byte range using RAND_MAX */
    bytebuf[i] = (unsigned char)((255 * rand())/RAND_MAX);
  }
}

/* Note: Windows implementation just returns
   NULL, in order to return something*/
struct proc_dir_entry *proc_net_create(const char *name,
				       mode_t mode, get_info_t *get_info) {
  return NULL;
}

void proc_net_remove(const char *name) {
  /* Empty implementation in Windows version */
}

unsigned short ip_fast_csum(unsigned char *iph, unsigned int ihl) {
  __u32 csum = 0;
  unsigned int i;

  /* Make 16 bit words out of every two adjacent 8 bit words
     in the packet and add them up */
  for (i = 0; i < ihl; i+=2) {
    csum += (__u32)(((iph[i]<<8) & 0xff00) + (iph[i + 1] & 0xff));
  }

  /* Take only 16 bits out of the 32 bit sum */
  while (csum >> 16) {
    csum = (csum & 0xffff) + (csum >> 16);
  }

  /* One's complement the result */
  return (__u16)~csum;
}

/* Function used to retrieve a single random byte */
unsigned char getRandomByte() {
  unsigned char b;

  get_random_bytes(&b, 1);
  return b;
}

/* Function used to return the number of milliseconds
   passed since January 1, 1601 */
/* Note: Precision is lost here on Windows; May lead
   to some timer glitches; TODO: Improve the solution */
unsigned long get_millis() {
  LARGE_INTEGER cur_time; /* LARGE_INTEGER 64 bits */
  __u64 ret_time;

  /* This function returns 64 bit time value */
  NdisGetCurrentSystemTime((PLARGE_INTEGER)&cur_time);

  ret_time = (__u64)((((__u64)cur_time.HighPart << 32) & 0xffffffff00000000) +
		     ((__u64)cur_time.LowPart & 0x00000000ffffffff));

  /* The value returned is in 100 ns */
  return (unsigned long)(ret_time / 10);
}

/* Function used to initialize some functionality
   provided by the platform */
void platform_init() {
  /* Initialize random number generator */

  /* TODO: Perhaps use time.h instead */
  /* srand((unsigned)time(NULL)); */

  unsigned long cur_time = get_millis();
  /* Get lower 32 bits and cast to an unsigned (32 bit) int */
  srand((__u32)cur_time);
}

/* Function used to finalize some functionality
   provided by the platform */
void platform_fini() {
}
